"""
Integration with Jupyter notebooks. This module provides a
:class:`~LineageBuilder` subclass to simplify Lineage for Notebooks.

It also provides a collection of IPython *magics* (macros) for working
in Jupyter notebooks.
"""
import os
import sys
import ipykernel # type: ignore
from IPython.core.getipython import get_ipython # type: ignore
from IPython.core.magic import (Magics, magics_class, line_magic) # type: ignore
from IPython.core.display import display # type: ignore
from IPython.display import IFrame, HTML, Markdown # type: ignore

import requests
import json
from urllib.parse import urljoin
import re
from os.path import join, basename, dirname, abspath, expanduser, curdir, exists
from notebook.notebookapp import list_running_servers # type: ignore
from typing import Optional, List, Any, Dict, Tuple, Callable, Sequence
assert Dict # keep pyflakes happy
import shlex
import argparse

from collections import namedtuple


from dataworkspaces.lineage import LineageBuilder
from dataworkspaces.workspace import _find_containing_workspace
from dataworkspaces.api import take_snapshot, get_snapshot_history,\
                               make_lineage_table, make_lineage_graph,\
                               get_results, get_resource_info
from dataworkspaces.errors import ConfigurationError


def _get_notebook_name(verbose=False) -> Optional[str]:
    """
    Return the full path of the jupyter notebook.
    See https://github.com/jupyter/notebook/issues/1000

    In some situations (e.g. running on the command line via nbconvert),
    the notebook name is not available. We return None in those cases.
    """
    # kernel_id = re.search('kernel-(.*).json',
    #                       ipykernel.connect.get_connection_file()).group(1)
    try:
        ipy = get_ipython()
        info = ipy.ev("DWS_JUPYTER_INFO")
        return info.notebook_path
    except Exception as e:
        if verbose:
            print("DWS Jupyter extension was not loaded: %s" % e)
    try:
        connection_file = ipykernel.connect.get_connection_file()
        mo=re.search('kernel-(.*).json', connection_file)
        if mo is not None:
            kernel_id = mo.group(1)
            servers = list_running_servers()
            for ss in servers:
                response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                        params={'token': ss.get('token', '')})
                for nn in json.loads(response.text):
                    if nn['kernel']['id'] == kernel_id:
                        relative_path = nn['notebook']['path']
                        return join(ss['notebook_dir'], relative_path)
            print("Did not find a matching notebook server for %s" % connection_file)
            return None
    except Exception as e:
        if verbose:
            print("Unable to use notebook API to access session info: %s" % e)
    # all our atempts failed
    return None


def _remove_notebook_extn(notebook_name):
    if notebook_name.endswith('.ipynb'):
        return notebook_name[0:-6]
    elif notebook_name.endswith('.py'):
        return notebook_name[0:-3]
    else:
        return notebook_name

def get_step_name_for_notebook() -> Optional[str]:
    """
    Get the step name for a notebook by getting the path and then
    extracting the base name.
    In some situations (e.g. running on the command line via nbconvert),
    the notebook name is not available. We return None in those cases.
    """
    notebook_path = _get_notebook_name()
    if notebook_path is not None:
        return _remove_notebook_extn(basename(notebook_path))
    else:
        return None


def is_notebook() -> bool:
    """Return true if this code is running in a notebook.
    """
    try:
        # if running in ipython, get_ipython() will be in the global contect
        shell = get_ipython().__class__.__name__ # type: ignore
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter or a script


def get_notebook_directory():
    notebook_path = _get_notebook_name()
    if notebook_path is not None:
        return dirname(notebook_path)
    else:
        return curdir


class NotebookLineageBuilder(LineageBuilder):
    """Notebooks are the final step in a pipeline
    (and potentially the only step). We customize
    the standard lineage builder to get the step
    name from the notebook's name and to always have
    a results directory.

    If you are not running this notebook in a server
    context (e.g. via nbconvert), the step name won't be
    available. In that case, you can explicitly pass in the
    step name to the constructor.
    """
    def __init__(self, results_dir:str,
                 step_name:Optional[str]=None,
                 run_description:Optional[str]=None):
        super().__init__()
        if step_name is not None:
            self.step_name = step_name
        else:
            notebook_step_name = get_step_name_for_notebook()
            if notebook_step_name is None:
                raise ConfigurationError("Unable to infer the name of this notebook. "+
                                         "Please either use the DWS notebook magic or pass the name in explicitly to the lineage builder.")
            self.step_name = notebook_step_name
        notebook_path = _get_notebook_name()
        if notebook_path is not None:
            self.code.append(notebook_path)
        else:
            # if we are not running in a server content,
            # use the current directory as the code resource path
            self.code.append(abspath(curdir))
        self.results_dir = results_dir
        self.run_description = run_description


############################################################################
#                 Code for IPython magic extension                         #
############################################################################

DwsJupyterInfo = namedtuple('DwsJupyterInfo',
                            ['notebook_name', 'notebook_path', 'workspace_dir', 'notebook_server_dir',
                             'error'])



init_jscode=r"""%%javascript
var dws_initialization_msg = "Ran DWS initialization. The following magic commands have been added to your notebook:\n- `%dws_info` - print information about your dws environment\n- `%dws_history` - print a history of snapshots in this workspace\n- `%dws_snapshot` - save and create a new snapshot\n- `%dws_lineage_table` - show a table of lineage for the workspace resources\n- `%dws_lineage_graph` - show a graph of lineage for a resource\n- `%dws_results` - show results from a run (results.json file)\n\nRun any command with the `--help` option to see a list\nof options for that command.\n\nThe variable `DWS_JUPYTER_NOTEBOOK` has been added to\nyour variables, for use in future DWS calls.\n\nIf you want to disable the DWS magic commands (e.g. when running in a batch context), set the variable `DWS_MAGIC_DISABLE` to `True` ahead of the `%load_ext` call.";
if (typeof Jupyter == "undefined") {
    alert("Unable to initialize DWS magic. This version only works with Jupyter Notebooks, not nbconvert or JupyterLab.");
    throw "Unable to initialize DWS magic. This version only works with Jupyter Notebooks, not nbconvert or JupyterLab.";
}
else if (Jupyter.notebook.hasOwnProperty('kernel') && Jupyter.notebook.kernel!=null) {
    var DWSComm = Jupyter.notebook.kernel.comm_manager.new_comm('dws_comm_target', {})
    DWSComm.on_msg(function(msg) {
        console.log("DWS Got msg status: " + msg.content.data.status);
        console.log("DWS msg type: " + msg.content.data.msg_type);
        if (msg.content.data.status != 'ok') {
            if (msg.content.data.hasOwnProperty('cell')) {
                var cell = Jupyter.notebook.get_cell(msg.content.data.cell-1);
                cell.output_area.append_output({'data':{'text/plain':msg.content.data.status}, 'metadata':{}, 'output_type':'display_data' });
            }
            alert(msg.content.data.status);
            return;
        }
        if (msg.content.data.msg_type == "snapshot-result") {
            var cell = Jupyter.notebook.get_cell(msg.content.data.cell-1);
            cell.output_area.append_output({'data':{'text/plain':msg.content.data.message}, 'metadata':{}, 'output_type':'display_data' });
            alert(msg.content.data.message);
        }
        else if (msg.content.data.msg_type == "init-ack") {
            var cell = Jupyter.notebook.get_cell(msg.content.data.cell-1);
            cell.output_area.append_output({'data':{'text/markdown':dws_initialization_msg}, 'metadata':{}, 'output_type':'display_data' });
            //alert(dws_initialization_msg);
        }
    });
    // Send data
    var cellno = Jupyter.notebook.find_cell_index(Jupyter.notebook.get_selected_cell());
    DWSComm.send({'msg_type':'init',
                  'notebook_name': Jupyter.notebook.notebook_name,
                  'notebook_path': Jupyter.notebook.notebook_path,
                  'cell':cellno});
    window.DWSComm = DWSComm;
} else {
    // this happens when evaluating the javascript upon loading the notebook
    console.log("kernal was null");
}
"""

snapshot_jscode="""%%javascript
Jupyter.notebook.save_notebook();
"""

snapshot_jscode2="""%%javascript
if (window.hasOwnProperty('DWSComm')) {
    window.DWSComm.send({'msg_type':'snapshot',
                         'cell':Jupyter.notebook.find_cell_index(Jupyter.notebook.get_selected_cell())});
    console.log("sending snapshot");
}
"""


class DwsMagicError(ConfigurationError):
    pass

class DwsMagicArgParseExit(Exception):
    """Throw this in our overriding the exit method"""
    pass

class DwsMagicParseArgs(argparse.ArgumentParser):
    """Specialized version of the argument parser that can
    work in the context of ipython magic commands. Should
    never call sys.exit() and needs its own line parsing.
    """
    def parse_magic_line(self, line):
        return self.parse_args(shlex.split(line))
    def error(self, msg):
        raise DwsMagicError(msg)
    def exit(self, status=0, message=None):
        assert status==0, "Expecting a status of 0"
        raise DwsMagicArgParseExit()

# Colormaps for heatmaps
# These were generated using seaborn:
#     def tobyte(c):
#         return int(round(255*c))
#     MINIMIZE_COLORMAP = ['rgb(%s,%s,%s)'%(tobyte(c[0]),tobyte(c[1]),tobyte(c[2]))
#                          for c in seaborn.diverging_palette(150, 10, s=50, l=50, n=7)]
#     MAXIMIZE_COLORMAP = ['rgb(%s,%s,%s)'%(tobyte(c[0]),tobyte(c[1]),tobyte(c[2]))
#                          for c in seaborn.diverging_palette(10, 150, s=50, l=50, n=7)]
# The two maps are just the reverse of each other with maximize having greener colors toward
# the high indexes and redder colors toward the low indexes, and minimize being the opposite.
# By pre-generating the colormaps, we avoid a dependency on seaborn.
MINIMIZE_COLORMAP=['rgb(84,128,107)', 'rgb(138,168,153)', 'rgb(193,210,201)', 'rgb(242,242,242)', 'rgb(232,190,192)', 'rgb(212,136,140)', 'rgb(193,84,89)']
MAXIMIZE_COLORMAP=['rgb(193,84,89)', 'rgb(212,136,140)', 'rgb(232,190,192)', 'rgb(242,242,242)', 'rgb(193,210,201)', 'rgb(138,168,153)', 'rgb(84,128,107)']

def _fmt_scalar(s):
    """Helper function to round metrics"""
    if not isinstance(s, float) and (not hasattr(s, 'dtype') or s.dtype!='f'):
        return s # non-floats left alone
    elif s >=1.0:
        return round(s, 1)
    elif s>=0.01:
        return round(s, 3)
    else:
        return s

_BINS_TO_LABELS={
  1: [3],
  2: [2,4],
  3: [2,3,4],
  4: [1,2,4,5],
  5: [1,2,3,4,5],
  6: [0,1,2,4,5,6],
  7: [0,1,2,3,4,5,6]
} # type: Dict[int,Sequence[int]]
def _metric_col_to_colormap(col):
    """Given a metric column, return a series representing
    the heatmap indexes (0 through 6).
    Returns a series with the same number of elements as the column.
    """
    import pandas as pd # type: ignore
    import numpy as np # type: ignore
    nunique = len(col.dropna().unique())
    num_bins = min(nunique, 7)
    if num_bins<2:
        return col.apply(lambda v: -1 if pd.isna(v) else 3)
    elif num_bins==2:
        minval=col.min()
        return col.apply(lambda v: -1 if pd.isna(v)
                         else (2 if v==minval else 4))
    # qcut() may collapse bins, so we need to figure out how many bins it will
    # actually give us.
    num_actual_bins = len(pd.qcut(col, num_bins, duplicates='drop').dtype.categories)
    labels=_BINS_TO_LABELS[num_actual_bins]
    try:
        return pd.qcut(col, num_bins, labels=labels, duplicates='drop').astype(np.float32).fillna(-1.0).astype(np.int32)
    except Exception as e:
        print(e, file=sys.stderr)
        print("problem binning columns, unique=%s, num_actual_bins=%s, labels=%s"%
              (nunique, num_actual_bins, labels), file=sys.stderr)
        print("col: %s" % repr(col), file=sys.stderr)
        raise


@magics_class
class DwsMagics(Magics):
    def __init__(self, shell):
        super().__init__(shell)
        try:
            self.disabled = get_ipython().ev('DWS_MAGIC_DISABLE')
        except NameError:
            self.disabled = False
        if self.disabled:
            print("Loaded Data Workspaces magic commands in disabled state.", file=sys.stderr)
            return
        self._snapshot_args = None # type: Optional[argparse.Namespace] 
        def target_func(comm, open_msg):
            self.comm = comm
            @comm.on_msg
            def _recv(msg):
                ipy = get_ipython()
                data = msg['content']['data']
                msg_type = data['msg_type']
                if msg_type=='init':
                    # It looks like the notebook is always running with the cwd set to the notebook
                    # However, the notebook path from the browser is relative to where the
                    # notebook server was started
                    #npath = data['notebook_path']
                    #if not isabs(npath):
                    #    npath = join(abspath(expanduser(curdir)), npath)
                    abscwd = abspath(expanduser(os.getcwd()))
                    npath = join(abscwd, data['notebook_name'])
                    assert exists(npath), "Wrong calculation for absolute notebook path, got %s" % npath
                    assert npath.endswith(data['notebook_path']), \
                        "Unexpacted notebook path from client, got %s, but absolute is %s" %\
                        (data['notebook_path'], npath)
                    notebook_server_dir = npath[0:-(len(data['notebook_path'])+1)]
                    notebook_dir=dirname(npath)
                    workspace_dir = _find_containing_workspace(notebook_dir)
                    error = None
                    if workspace_dir is None:
                        error = "Unable to find a containing workspace for note book at %s" % npath
                    DWS_JUPYTER_INFO=DwsJupyterInfo(data['notebook_name'],
                                                    npath,
                                                    workspace_dir,
                                                    notebook_server_dir,
                                                    error)
                    ipy.push({'DWS_JUPYTER_INFO': DWS_JUPYTER_INFO})
                    if error:
                        comm.send({'status':error})
                        raise Exception(error)
                    else:
                        comm.send({'status':'ok', 'msg_type':'init-ack', 'cell':data['cell']})
                        self.dws_jupyter_info = DWS_JUPYTER_INFO
                elif msg_type=='snapshot':
                    cell = data['cell']
                    try:
                        assert self._snapshot_args is not None
                        r = take_snapshot(self.dws_jupyter_info.workspace_dir,
                                          tag=self._snapshot_args.tag,
                                          message=self._snapshot_args.message)
                        self._snapshot_args = None
                        comm.send({'msg_type':'snapshot-result',
                                   'status':'ok',
                                   'message':'Successfully completed snapshot. Hash is %s'%r[0:8],
                                   'cell':cell})
                    except Exception as e:
                        comm.send({'msg_type':'snapshot-result',
                                   'status':"Snapshot failed with error '%s'"% e,
                                   'cell':cell})
                        raise
                else:
                    raise Exception("Unknown message type %s" % msg_type)
        self.shell.kernel.comm_manager.register_target('dws_comm_target', target_func)
        self.shell.run_cell(init_jscode, store_history=False, silent=True)

    async def _call_snapshot(self):
        await self.shell.run_cell_async(snapshot_jscode)
        await self.shell.run_cell_async(snapshot_jscode2)

    @line_magic
    def dws_info(self, line):
        import pandas as pd # TODO: support case where pandas wasn't installed
        parser = DwsMagicParseArgs("dws_info",
                                   description="Print some information about this workspace")
        try:
            parser.parse_magic_line(line)
        except DwsMagicArgParseExit:
            return # user asked for help
        if self.disabled:
            display(Markdown("DWS magic commands are disabled. To enable, set `DWS_MAGIC_DISABLE` to `False` and restart kernel."))
            return
        print("Notebook name:       %s" % self.dws_jupyter_info.notebook_name)
        print("Notebook path:       %s"  % self.dws_jupyter_info.notebook_path)
        print("Workspace directory: %s" % self.dws_jupyter_info.workspace_dir)
        print("Notebook server dir: %s" % self.dws_jupyter_info.notebook_server_dir)
        if self.dws_jupyter_info.error is not None:
            print("Error message:       %s" % self.dws_jupyter_info.error)
            return

        resources = get_resource_info(self.dws_jupyter_info.workspace_dir)
        df = pd.DataFrame({
            'Resource':[r.name for r in resources],
            'Role':[r.role for r in resources],
            'Type':[r.resource_type for r in resources],
            'Local Path':[r.local_path for r in resources]
        })
        with pd.option_context('display.max_colwidth', 80):
            display(df)

    @line_magic
    def dws_snapshot(self, line):
        parser = DwsMagicParseArgs("dws_snapshot",
                                   description="Save the notebook and create a new snapshot")
        parser.add_argument('-m', '--message', type=str, default=None,
                            help="Message describing the snapshot")
        parser.add_argument('-t', '--tag', type=str, default=None,
                            help="Tag for the snapshot. Note that a given tag can "+
                                 "only be used once (without deleting the old one).")
        try:
            args = parser.parse_magic_line(line)
        except DwsMagicArgParseExit:
            return # user asked for help
        if self.disabled:
            display(Markdown("DWS magic commands are disabled. To enable, set `DWS_MAGIC_DISABLE` to `False` and restart kernel."))
            return
        self._snapshot_args = args
        msg = "Initiating snapshot"
        if args.tag:
            msg += " with tag '%s'" % args.tag
        if args.message:
            msg += " with message '%s'" % args.message
        print(msg + '...')
        import tornado.ioloop
        tornado.ioloop.IOLoop.current().spawn_callback(self._call_snapshot)

    @line_magic
    def dws_history(self, line):
        parser = DwsMagicParseArgs("dws_history",
                                   description="Print a history of snapshots in this workspace")
        parser.add_argument('--max-count', type=int, default=None,
                            help="Maximum number of snapshots to show")
        parser.add_argument('--tail', default=False, action='store_true',
                            help="Just show the last 10 entries in reverse order")
        parser.add_argument('--baseline', default=None, type=str,
                            help="Snapshot tag or hash to use as a basis for metrics comparison. "+
                                 "Will color the fonts of values green or red, "+
                                 "depending on whether they are better (worse) than "+
                                 "the baseline.")
        parser.add_argument('--heatmap', default=False, action='store_true',
                            help="Show a heatmap for metrics columns")
        parser.add_argument('--maximize-metrics', default=None, type=str,
                            help="Metrics where larger values are better (e.g. accuracy)")
        parser.add_argument('--minimize-metrics', default=None, type=str,
                            help="Metrics where smaller values are better (e.g. loss)")
        # TODO: future feature
        # parser.add_argument('--round-metrics', type=int, default=None,
        #                     help="If specified, round metrics to this many decimal places")
        try:
            args = parser.parse_magic_line(line)
        except DwsMagicArgParseExit:
            return # user asked for help
        if self.disabled:
            display(Markdown("DWS magic commands are disabled. To enable, set `DWS_MAGIC_DISABLE` to `False` and restart kernel."))
            return
        import pandas as pd # TODO: support case where pandas wasn't installed
        import numpy as np # type: ignore
        if args.heatmap:
            if args.baseline is not None:
                print("Cannot specify both --baseline and --heatmap", file=sys.stderr)
                return
        if args.max_count and args.tail:
            max_count = args.max_count
        elif args.tail:
            max_count = 10
        else:
            max_count = None
        history = get_snapshot_history(self.dws_jupyter_info.workspace_dir,
                                       max_count=max_count,
                                       reverse=args.tail)
        entries = []
        index = []
        columns = ['timestamp', 'hash', 'tags', 'message']
        baseline_snapshot = None # type: Optional[int]
        # not every snapshot has the same metrics, so we build an inclusive list
        metrics = [] # type: List[str]
        for s in history:
            d = {'timestamp':s.timestamp[0:19],
                 'hash':s.hashval[0:8],
                 'tags':', '.join([tag for tag in s.tags]),
                 'message':s.message if s.message is not None else ''}
            if s.metrics is not None:
                for (m, v) in s.metrics.items():
                    d[m] = v
                    if m not in columns:
                        columns.append(m)
                        metrics.append(m)
            entries.append(d)
            index.append(s.snapshot_number)
            if (args.baseline is not None):
                if args.baseline in s.tags:
                    baseline_snapshot = s.snapshot_number
                elif s.hashval[0:min(len(args.baseline),8)]==args.baseline[0:8]:
                    baseline_snapshot = s.snapshot_number
        if (args.baseline is not None) and (baseline_snapshot is None):
            print("Did not find a tag or hash corresponding to baseline '%s'"
                  % args.baseline, file=sys.stderr)
            return
        history_df = pd.DataFrame(entries, index=index, columns=columns)
        maximize_metrics = set(['accuracy', 'precision', 'recall'])
        if args.maximize_metrics:
            maximize_metrics = maximize_metrics.union(set(args.maximize_metrics.split(',')))
        minimize_metrics = set(['loss'])
        if args.minimize_metrics:
            minimize_metrics = minimize_metrics.union(set(args.minimize_metrics.split(',')))
        def truncate(v, l=30):
            s = repr(v)
            return s if len(s)<=(l-3) else s[0:l-3]+'...'
        def cleanup_dict_or_string_metric(val):
            if isinstance(val, dict) or isinstance(val, str):
                return truncate(val)
            else:
                return val
        element_styling_fns = [] # type: List[Tuple[str, Callable[[Any], None]]]
        if args.heatmap:
            heatmap_maximize_cols = [] # type: List[str]
            heatmap_minimize_cols = [] # type: List[str]
            color_templ="border: 1px solid darkgrey; background-color: %s; color: %s"
            # TODO: split this out to a separate function
            def color_max_metric_col(col):
                bins = _metric_col_to_colormap(col)
                return bins.apply(lambda b: color_templ%(MAXIMIZE_COLORMAP[b], 'white' if b<2 or b>4 else  'black') if b!=-1
                                            else color_templ%('lightgrey', 'black'))
            def color_min_metric_col(col):
                bins = _metric_col_to_colormap(col)
                return bins.apply(lambda b: color_templ%(MINIMIZE_COLORMAP[b], 'white' if b<2 or b>4 else  'black') if b!=-1
                                            else color_templ%('lightgrey', 'black'))
        class BaselineElementStyle:
            def __init__(self, metric:str, baseline, maximize:bool):
                self.metric=metric
                self.baseline=baseline
                self.baseline_round = abs(self.baseline*0.005)
                self.maximize=maximize
            def __call__(self, val):
                # if a value is within 0.5% of the baseline, we consider it baseline
                if pd.isna(val):
                    return 'color: grey'
                elif val>(self.baseline+self.baseline_round):
                    return 'color: green' if self.maximize else 'color: red'
                elif val<(self.baseline-self.baseline_round):
                    return 'color: red' if self.maximize else 'color: green'
                else: # within baseline rounding
                    return 'color: black; font-weight: bold'
        for metric in metrics:
            if history_df[metric].dtype.kind in ('f', 'i'):
                # float or int
                if baseline_snapshot is not None:
                    baseline_val = history_df.loc[baseline_snapshot][metric]
                    if metric in maximize_metrics:
                        element_styling_fns.append((metric, BaselineElementStyle(metric, baseline_val, maximize=True)),)
                    elif metric in minimize_metrics:
                        element_styling_fns.append((metric, BaselineElementStyle(metric, baseline_val, maximize=False)),)
                elif args.heatmap:
                    if metric in maximize_metrics:
                        heatmap_maximize_cols.append(metric)
                    elif metric in minimize_metrics:
                        heatmap_minimize_cols.append(metric)
            elif history_df[metric].dtype==np.dtype('object'):
                history_df[metric] = history_df[metric].apply(cleanup_dict_or_string_metric)
        result = history_df
        def get_style(df_or_style):
            return df_or_style.style if isinstance(df_or_style, pd.DataFrame) else df_or_style
        for (metric, styling_fn) in element_styling_fns:
            result = get_style(result).applymap(styling_fn, subset=[metric])
        if args.heatmap:
            result = get_style(result).apply(color_max_metric_col, subset=heatmap_maximize_cols)
            result = get_style(result).apply(color_min_metric_col, subset=heatmap_minimize_cols)
        return result

    @line_magic
    def dws_lineage_table(self, line):
        import pandas as pd # TODO: support case where pandas wasn't installed
        parser = DwsMagicParseArgs("dws_lineage_table",
                                   description="Show a table of lineage for the workspace's resources")
        parser.add_argument('--snapshot', default=None, type=str,
                            help="If specified, print lineage as of the specified snapshot hash or tag")
        try:
            args = parser.parse_magic_line(line)
        except DwsMagicArgParseExit:
            return # user asked for help
        if self.disabled:
            display(Markdown("DWS magic commands are disabled. To enable, set `DWS_MAGIC_DISABLE` to `False` and restart kernel."))
            return
        rows = [r for r in make_lineage_table(self.dws_jupyter_info.workspace_dir, args.snapshot)]
        return pd.DataFrame(rows, columns=['Resource', 'Lineage Type', 'Details', 'Inputs']).set_index('Resource')

    @line_magic
    def dws_lineage_graph(self, line):
        parser = DwsMagicParseArgs("dws_lineage_table",
                                   description="Show a graph of lineage for a resource")
        parser.add_argument('--resource', default=None, type=str,
                            help="Graph lineage from this resource. Defaults to the results resource. Error if not specified and there is more than one.")
        parser.add_argument('--snapshot', default=None, type=str,
                            help="If specified, graph lineage as of the specified snapshot hash or tag")
        try:
            args = parser.parse_magic_line(line)
        except DwsMagicArgParseExit:
            return # user asked for help
        if self.disabled:
            display(Markdown("DWS magic commands are disabled. To enable, set `DWS_MAGIC_DISABLE` to `False` and restart kernel."))
            return
        output_file = join(dirname(self.dws_jupyter_info.notebook_path),
                           'lineage_'+_remove_notebook_extn(self.dws_jupyter_info.notebook_name)+'.html')
        make_lineage_graph(output_file, self.dws_jupyter_info.workspace_dir,
                           resource_name=args.resource, tag_or_hash=args.snapshot,
                           width=780, height=380)
        return display(IFrame(basename(output_file), width=800, height=400))

    @line_magic
    def dws_results(self, line):
        parser = DwsMagicParseArgs("dws_results",
                                   description="Show results from a run (results.json file)")
        parser.add_argument('--resource', default=None, type=str,
                            help="Look for the results.json file in this resource. Otherwise, will look in all results resources and return the first match.")
        parser.add_argument('--snapshot', default=None, type=str,
                            help="If specified, get results as of the specified snapshot or tag. Otherwise, looks at current workspace and then most recent snapshot.")
        try:
            args = parser.parse_magic_line(line)
        except DwsMagicArgParseExit:
            return # user asked for help
        if self.disabled:
            display(Markdown("DWS magic commands are disabled. To enable, set `DWS_MAGIC_DISABLE` to `False` and restart kernel."))
            return
        rtn = get_results(self.dws_jupyter_info.workspace_dir,
                          tag_or_hash=args.snapshot, resource_name=args.resource)
        if rtn is None:
            print("Did not find a results.json file.", file=sys.stderr)
            return
        (results, rpath) = rtn
        import pandas as pd
        html_list = ['<h3>%s</h3>' % rpath]

        def truncate_dict(d, maxlen=50, roundme=False):
            d2 = {}
            for (k, v) in d.items():
                if roundme:
                    d2[k] = _fmt_scalar(v)
                else:
                    d2[k] = v
            s = repr(d2)
            if len(s)>maxlen:
                return s[0:(maxlen-3)]+'...'
            else:
                return s
        def subdict_to_df(d, parent_name, name, roundme=False):
            keys=[]
            values = []
            for (k, v) in d.items():
                if not isinstance(v, dict):
                    keys.append(k)
                    if roundme:
                        values.append(_fmt_scalar(v))
                    else:
                        values.append(v)
                else:
                    keys.append(k)
                    values.append(truncate_dict(v, roundme=roundme))
            df = pd.DataFrame({'Property':keys, 'Value':values}).set_index('Property')
            html_list.append("<h5>%s: %s</h5>"% (parent_name, name))
            html_list.append(df.to_html())
        def dict_to_df(d, name, roundme=False):
            keys=[]
            values = []
            subdicts = []
            for (k, v) in d.items():
                if not isinstance(v, dict):
                    keys.append(k)
                    if roundme:
                        values.append(_fmt_scalar(v))
                    else:
                        values.append(v)
                elif k not in ('parameters', 'metrics'):
                    subdicts.append((k, v))
            df = pd.DataFrame({'Property':keys, 'Value':values}).set_index('Property')
            html_list.append("<h4>%s</h4>"% name)
            html_list.append(df.to_html())
            for (k, v) in subdicts:
                subdict_to_df(v, name, k, roundme=roundme)
        dict_to_df(results, 'General Properties')
        if 'parameters' in results:
            dict_to_df(results['parameters'], 'Parameters')
        if 'metrics' in results:
            dict_to_df(results['metrics'], 'Metrics', roundme=True)
        return HTML('\n'.join(html_list))



def load_ipython_extension(ipython):
    ipython.register_magics(DwsMagics)




# Copyright 2018 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Resource for files copied in by rclone 
"""
import os
import os.path

# import stat
from typing import List, Optional, Any, cast
import json
import click

from dataworkspaces.errors import ConfigurationError
from dataworkspaces.workspace import (
    Workspace,
    LocalStateResourceMixin,
    SnapshotWorkspaceMixin,
    JSONDict,
    ResourceFactory,
)
from dataworkspaces.utils.file_utils import LocalPathType
from dataworkspaces.third_party.rclone import RClone
from dataworkspaces.utils.param_utils import (
    ParamType,
    ParamValidationError,
    StringType,
    EnumType,
    BoolType,
)

from dataworkspaces.resources.local_file_resource import LocalFileResource, setup_path_for_hashes

RCLONE_RESOURCE_TYPE = "rclone"

"""
dws add rclone [options] remote local

See 
"""


class RemoteOriginType(ParamType):
    """Custom param type for maintaining the remote origin in the form
    remote_name:path. Path does not need to be absolute
    """

    def validate(self, value: Any) -> None:
        if not isinstance(value, str):
            raise ParamValidationError("Remote origin '%s' is not a string" % repr(value))
        if ":" not in value:
            raise ParamValidationError(
                "Remote origin '%s' is missing a ':', should be of the form remote:path" % value
            )

    def __str__(self):
        return "remote_origin"


class RcloneResource(LocalFileResource):
    """Local files synchronized to a remote via rclone."""

    def __init__(
        self,
        name: str,
        role: str,
        workspace: Workspace,
        remote_origin: str,
        global_local_path: str,
        my_local_path: Optional[str],
        config: Optional[str] = None,
        compute_hash: Optional[bool] = None,
        export: Optional[bool] = None,
        imported: Optional[bool] = None,
        master: Optional[str] = None,
        sync_mode: Optional[str] = None,
        size_only: Optional[bool] = None,
        ignore: List[str] = [],
    ):
        super().__init__(
            RCLONE_RESOURCE_TYPE,
            name,
            role,
            workspace,
            global_local_path,
            my_local_path,
            compute_hash,
            export,
            imported,
            ignore,
        )
        self.param_defs.define(
            "remote_origin",
            default_value=None,
            optional=False,
            is_global=True,
            help="Rclone remote origin in the form remote:path",
            ptype=RemoteOriginType(),
        )
        self.remote_origin = self.param_defs.get("remote_origin", remote_origin)  # type: str
        (self.remote_name, _) = self.remote_origin.split(":")
        self.param_defs.define(
            "config",
            default_value=None,
            optional=True,
            is_global=True,
            help="Optional path to rclone config file (otherwise uses the default)",
            ptype=StringType(),
        )
        self.config = self.param_defs.get("config", config)  # type: Optional[str]
        self.param_defs.define(
            "master",
            default_value="none",
            optional=False,
            help="Determines which system is the master. If 'remote', then pulls will be done, but not pushes. "
            + "If 'local', then pushes will be done, but not pulls. If 'none' (the default), no action will be taken "
            + "for pushes and pulls (you need to synchronize manually using rclone). When first adding the resource "
            + " or cloning to a new machine, if the local directory does not exist, and 'remote' or 'none' were specified, "
            + " the contents of the remote will copied down to the local directory.",
            is_global=True,
            ptype=EnumType("none", "remote", "local"),
        )
        self.master = self.param_defs.get("master", master)  # type: str
        self.param_defs.define(
            "sync_mode",
            default_value="copy",
            optional=False,
            help="When copying between local and master, which rclone command to use. If you specify 'copy', files are "
            + "added or overwritten without deleting any files present at the target. If you specify 'sync', files at "
            + "the target are removed if they are not present at the source. The default is 'copy'. If master is 'none', "
            + "this setting has no effect.",
            is_global=True,
            ptype=EnumType("sync", "copy"),
        )
        self.sync_mode = self.param_defs.get("sync_mode", sync_mode)  # type: str
        self.param_defs.define(
            "size_only",
            default_value=False,
            optional=False,
            help="If specified, use only the file size (rather than also modification time and checksum) to "
            + "determine if a file has been changed. If your resource has a lot of files and access to the remote "
            + "is over a WAN, you probably want to set this. Otherwise, syncs/copies can be VERY slow.",
            is_global=True,
            ptype=BoolType(),
            allow_missing=True,
        )
        self.size_only = self.param_defs.get("size_only", size_only)  # type: bool

        if config:
            self.rclone = RClone(cfgfile=self.config)
        else:
            self.rclone = RClone()

    def pull_precheck(self) -> None:
        pass  # a dry run can be very expensive, so skipping for now

    def pull(self) -> None:
        if self.size_only:
            flags = ["--size-only"]
        else:
            flags = []
        flags.append("--verbose")
        if self.master == "remote":
            if self.sync_mode == "sync":
                ret = self.rclone.sync(self.remote_origin, self.local_path, flags=flags)
                if ret["code"] != 0:
                    raise ConfigurationError(
                        "rclone sync raised error %d: %s" % (ret["code"], ret["error"])
                    )
            elif self.sync_mode == "copy":
                ret = self.rclone.copy(self.remote_origin, self.local_path, flags=flags)
                if ret["code"] != 0:
                    raise ConfigurationError(
                        "rclone copy raised error %d: %s" % (ret["code"], ret["error"])
                    )
        else:
            click.echo("Skipping pull of resource %s, master is %s" % (self.name, self.master))

    def push_precheck(self) -> None:
        pass  # a dry run can be very expensive, so skipping for now

    def push(self) -> None:
        if self.size_only:
            flags = ["--size-only"]
        else:
            flags = []
        flags.append("--verbose")
        if self.master == "local":
            if self.sync_mode == "copy":
                ret = self.rclone.copy(self.local_path, self.remote_origin, flags=flags)
                if ret["code"] != 0:
                    raise ConfigurationError(
                        "rclone copy raised error %d: %s" % (ret["code"], ret["error"])
                    )
            else:
                ret = self.rclone.sync(self.local_path, self.remote_origin, flags=flags)
                if ret["code"] != 0:
                    raise ConfigurationError(
                        "rclone sync raised error %d: %s" % (ret["code"], ret["error"])
                    )
        else:
            click.echo("Skiping push of resource %s, master is %s" % (self.name, self.master))

    def __str__(self):
        return "Rclone-d repo %s, locally copied in %s in role '%s'" % (
            self.remote_origin,
            self.local_path,
            self.role,
        )


class RcloneFactory(ResourceFactory):
    def _add_prechecks(self, local_path, remote_origin, config) -> RClone:
        if os.path.exists(local_path) and not (os.access(local_path, os.W_OK)):
            raise ConfigurationError(local_path + " does not have write permission")
        if config:
            rclone = RClone(cfgfile=config)
        else:
            rclone = RClone()
        known_remotes = rclone.listremotes()
        (remote_name, _) = remote_origin.split(":")
        if remote_name not in known_remotes:
            raise ConfigurationError("Remote '" + remote_name + "' not found by rclone")
        return rclone

    def _copy_from_remote(
        self,
        name,
        local_path,
        remote_origin,
        rclone,
        master="none",
        sync_mode="copy",
        size_only=False,
        verbose=False,
    ):
        if master == "remote":
            click.echo("%s: performing initial %s from remote" % (name, sync_mode))
            if not os.path.exists(local_path):
                os.makedirs(local_path)
        elif master == "none" and (not os.path.exists(local_path)):
            click.echo("%s: performing initial copy from remote" % name)
            sync_mode = "copy"
            os.makedirs(local_path)
        elif master == "local" and not (os.path.exists(local_path)):
            click.echo("Creating empty directory at %s for resource %s" % (local_path, name))
            os.makedirs(local_path)
            return
        else:
            click.echo(
                "Skipping initial copy of resource %s to %s: path already exists and remote is local"
                % (name, local_path)
            )
            return

        if size_only:
            flags = ["--size-only"]
        else:
            flags = []
        flags.append("--verbose")

        if sync_mode == "copy":
            ret = rclone.copy(remote_origin, local_path, flags=flags)
        else:
            ret = rclone.sync(remote_origin, local_path, flags=flags)
        if ret["code"] != 0:
            raise ConfigurationError(
                "rclone %s raised error %d: %s" % (sync_mode, ret["code"], ret["error"])
            )
        # For now, leave the local copy alone
        # if master=='none':
        #     print("Marking files as readonly")
        #     for (dirpath, dirnames, filenames) in os.walk(local_path):
        #         for f_name in filenames:
        #             abspath = os.path.abspath(os.path.join(dirpath, f_name))
        #             mode = os.stat(abspath)[stat.ST_MODE]
        #             os.chmod(abspath, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)

    def from_command_line(
        self,
        role,
        name,
        workspace,
        remote_origin,
        local_path,
        config,
        compute_hash,
        export,
        imported,
        master,
        sync_mode,
        size_only,
    ):
        rclone = self._add_prechecks(local_path, remote_origin, config)
        self._copy_from_remote(
            name, local_path, remote_origin, rclone, master, sync_mode, size_only, workspace.verbose
        )
        setup_path_for_hashes(role, name, workspace, local_path)
        if imported:
            lineage_path = os.path.join(local_path, "lineage.json")
            if not os.path.exists(lineage_path):
                raise ConfigurationError(
                    "--imported was specified, but missing exported lineage file %s" % lineage_path
                )
            if (
                not isinstance(workspace, SnapshotWorkspaceMixin)
                or not workspace.supports_lineage()
            ):
                raise ConfigurationError(
                    "--imported was specified, but this workspace does not support lineage"
                )
            with open(lineage_path, "r") as f:
                lineage_data = json.load(f)
            if lineage_data["resource_name"] != name:
                raise ConfigurationError(
                    "Resource name in imported lineage '%s' does not match '%s'"
                    % (lineage_data["resource_name"], name)
                )
            cast(SnapshotWorkspaceMixin, workspace).get_lineage_store().import_lineage_file(
                name, lineage_data["lineages"]
            )

        return RcloneResource(
            name,
            role,
            workspace,
            remote_origin,
            global_local_path=local_path,
            my_local_path=None,
            config=config,
            compute_hash=compute_hash,
            export=export,
            imported=imported,
            master=master,
            sync_mode=sync_mode,
            size_only=size_only,
        )

    def from_json(
        self, params: JSONDict, local_params: JSONDict, workspace: Workspace
    ) -> RcloneResource:
        """Instantiate a resource object from the parsed resources.json file"""
        assert params["resource_type"] == RCLONE_RESOURCE_TYPE
        return RcloneResource(
            params["name"],
            params["role"],
            workspace,
            params["remote_origin"],
            # for backward compatibility, we also check for "local_path"
            global_local_path=params["global_local_path"]
            if "global_local_path" in params
            else params["local_path"],
            my_local_path=local_params["my_local_path"]
            if "my_local_path" in local_params
            else (local_params["local_path"] if "local_path" in local_params else None),
            config=params["config"],
            compute_hash=params["compute_hash"],
            export=params.get("export", False),
            imported=params.get("imported", False),
            master=params.get("master", None),
            sync_mode=params.get("sync_mode", None),
            size_only=params.get("size_only", None),
        )

    def has_local_state(self) -> bool:
        return True

    def clone(self, params: JSONDict, workspace: Workspace) -> LocalStateResourceMixin:
        """Instantiate a resource that was created remotely. In this case, we will
        copy from the remote origin.
        """
        remote_origin = params["remote_origin"]
        config = params["config"]
        name = params["name"]
        master = params.get("master", None)
        sync_mode = params.get("sync_mode", None)
        size_only = params.get("size_only", None)

        # check local_path, too for backward compatibility
        global_local_path = (
            params["global_local_path"] if "global_local_path" in params else params["local_path"]
        )  # type: str
        if os.path.exists(global_local_path):
            local_path = global_local_path
            my_local_path = None  # type: Optional[str]
        else:
            if not workspace.batch:
                # TODO: consider whether we can just create the directory the user specifies rather than
                # requiring it to pre-exist (since we will download anyway).
                local_path = cast(
                    str,
                    click.prompt(
                        "Rclone resource '%s' was located at '%s' on the original system. Where is it located on this system?"
                        % (name, global_local_path),
                        type=LocalPathType(exists=True),
                    ),
                )
                my_local_path = local_path
            else:
                raise ConfigurationError(
                    "Local files resource %s is missing from %s." % (name, global_local_path)
                )

        rclone = self._add_prechecks(local_path, remote_origin, config)
        self._copy_from_remote(
            name, local_path, remote_origin, rclone, master, sync_mode, size_only, workspace.verbose
        )
        return RcloneResource(
            name,
            params["role"],
            workspace,
            remote_origin,
            global_local_path=global_local_path,
            my_local_path=my_local_path,
            config=config,
            compute_hash=params["compute_hash"],
            export=params.get("export", False),
            imported=params.get("imported", False),
            master=master,
            sync_mode=sync_mode,
            size_only=size_only,
        )

    def suggest_name(
        self,
        workspace,
        role,
        remote_origin,
        local_path,
        config,
        compute_hash,
        export,
        imported,
        master,
        sync_mode,
        size_only,
    ):
        return os.path.basename(local_path)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
# import os
# from os.path import isabs, exists, expanduser, basename, dirname, isdir, join
# import shutil
# import json
# import datetime
# import subprocess

from typing import List

from dataworkspaces.workspace import Workspace
from dataworkspaces.errors import InternalError

# import dataworkspaces.commands.actions as actions
# from dataworkspaces.errors import ConfigurationError
# from dataworkspaces.utils.lineage_utils import get_current_lineage_dir

# EXECUTABLES_TO_EXCLUDE_FROM_STEP_NAME = ['python', 'python3', 'python2']

# class RemovePreviousLineage(actions.Action):
#     def __init__(self, ns, verbose, lineage_file):
#         super().__init__(ns, verbose)
#         self.lineage_file = lineage_file

#     def run(self):
#         if exists(self.lineage_file):
#             os.remove(self.lineage_file)

#     def __str__(self):
#         return "Remove old lineage file '%s', if present" % self.lineage_file


# class RunCommand(actions.Action):
#     @actions.provides_to_ns('start_ts', datetime.datetime)
#     @actions.provides_to_ns('end_ts', datetime.datetime)
#     def __init__(self, ns, verbose, command_and_args, cwd):
#         super().__init__(ns, verbose)
#         self.command_and_args = command_and_args
#         self.cwd = cwd

#     def run(self):
#         self.ns.start_ts = datetime.datetime.now()
#         if self.verbose:
#             click.echo(" ".join(self.command_and_args + "[run in %s]"%self.cwd))
#         cp = subprocess.run(self.command_and_args, cwd=self.cwd, encoding='utf-8')
#         cp.check_returncode()
#         self.ns.end_ts = datetime.datetime.now()

#     def __str__(self):
#         return "Run %s from directory %s" % (self.command_and_args, self.cwd)


# class WriteLineage(actions.Action):
#     @actions.requires_from_ns('start_ts', datetime.datetime)
#     @actions.requires_from_ns('end_ts', datetime.datetime)
#     def __init__(self, ns, verbose, lineage_file, lineage_data):
#         super().__init__(ns, verbose)
#         self.lineage_file = lineage_file
#         self.lineage_data = lineage_data

#     def run(self):
#         self.lineage_data['timestamp'] = self.ns.start_ts.isoformat()
#         self.lineage_data['elapsed_time_seconds'] = \
#             round((self.ns.end_ts-self.ns.start_ts).total_seconds(), 1)
#         parent_dir = dirname(self.lineage_file)
#         if not isdir(parent_dir):
#             os.mkdir(parent_dir)
#         with open(self.lineage_file, 'w') as f:
#             json.dump(self.lineage_data, f, indent=2)

#     def __str__(self):
#         return "Write lineage data to %s" % self.lineage_file


# def remove_extension(fname):
#     try:
#         return fname[:fname.rindex('.')]
#     except ValueError:
#         return fname


def run_command(workspace: Workspace, step_name: str, cwd: str, command: str, args: List[str]):
    raise InternalError("Run command not yet supported")
    # XXX Need to port and finish implementing
    # ns = actions.Namespace()
    # plan = [ ]
    # # find the command executable
    # if isabs(command):
    #     if not exists(command):
    #         raise ConfigurationError("Command executable '%s' does not exist" % command)
    #     if not os.access(command, os.X_OK):
    #         raise ConfigurationError("Command '%s' is not executable" % command)
    #     command_path = command
    # else:
    #     command_path = shutil.which(command)
    #     if command_path is None:
    #         raise ConfigurationError("Could not find command '%s'" % command)
    # command_and_args = [command_path] + list(args)

    # cwd = expanduser(cwd)

    # # figure out the step name
    # if step_name is None:
    #     if basename(command_path) in EXECUTABLES_TO_EXCLUDE_FROM_STEP_NAME:
    #         step_name = remove_extension(basename(args[0]))
    #     else:
    #         step_name = remove_extension(basename(command_path))

    # lineage_file = join(get_current_lineage_dir(workspace_dir), '%s.json' % step_name)
    # lineage_data = {
    #     'step_name':step_name,
    #     'command_path':command_path,
    #     'args': args,
    #     'cwd': cwd,
    # }
    # plan.append(RemovePreviousLineage(ns, verbose, lineage_file))
    # plan.append(RunCommand(ns, verbose, command_and_args, cwd))
    # plan.append(WriteLineage(ns, verbose, lineage_file, lineage_data))
    # try:
    #     actions.run_plan(plan, "Run command with lineage", "run command with lineage", batch=batch, verbose=verbose)
    # except:
    #     # if we get an error, we need to wipe out the lineage file
    #     if exists(lineage_file):
    #         os.remove(lineage_file)
    #     raise



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.

from typing import Optional, cast

import click

from dataworkspaces.utils.hash_utils import is_a_git_hash, is_a_shortened_git_hash
from dataworkspaces.errors import ConfigurationError, UserAbort
from dataworkspaces.workspace import Workspace, SnapshotMetadata, SnapshotWorkspaceMixin


_CONF_MESSAGE = (
    "A snapshot with this hash already exists. Do you want to update "
    + "the message from '%s' to '%s'?"
)


def merge_snapshot_metadata(
    old: SnapshotMetadata, new: SnapshotMetadata, batch: bool
) -> SnapshotMetadata:
    """Merge two snapshot metadatas for when someone creates
    a snapshot without making changes. They might have
    added more tags or changed the message.
    """
    assert old.hashval == new.hashval
    tags = old.tags + [tag for tag in new.tags if tag not in old.tags]
    if (
        old.message != new.message
        and (new.message is not None)
        and (new.message != "")
        and (batch is False)
        and click.confirm(_CONF_MESSAGE % (old.message, new.message))
    ):  # type:ignore
        message = new.message
    else:
        message = old.message
    if old.tags == new.tags and old.message == new.message:
        raise ConfigurationError("No differences from previous snapshot, doing nothing.")
    click.echo("Snapshot %s already exists, updating metadata..." % old.hashval)
    return SnapshotMetadata(
        hashval=old.hashval,
        tags=tags,
        message=message,
        hostname=new.hostname,
        timestamp=old.timestamp,
        relative_destination_path=new.relative_destination_path,
        # The restore hash may have changed, even if the content did not.
        # E.g., in the git subdirectory, the restore hash reflects the hash of the overall
        # repo rather than just the subdirectory.
        restore_hashes=new.restore_hashes,
        metrics=new.metrics,
        updated_timestamp=new.timestamp,
    )


def snapshot_command(workspace: Workspace, tag: Optional[str] = None, message: str = "") -> str:
    if (tag is not None) and (is_a_git_hash(tag) or is_a_shortened_git_hash(tag)):
        raise ConfigurationError(
            "Tag '%s' looks like a git hash. Please pick something else." % tag
        )

    if not isinstance(workspace, SnapshotWorkspaceMixin):
        raise ConfigurationError("Workspace %s does not support snapshots." % workspace.name)
    mixin = cast(SnapshotWorkspaceMixin, workspace)
    # Remove existing tag if present
    if tag is not None:
        try:
            existing_tag_md = mixin.get_snapshot_by_tag(tag)  # type: Optional[SnapshotMetadata]
        except ConfigurationError:
            existing_tag_md = None
        if existing_tag_md is not None:
            msg = "Tag '%s' already exists for snapshot %s taken %s" % (
                tag,
                existing_tag_md.hashval,
                existing_tag_md.timestamp,
            )
            if workspace.batch:
                raise ConfigurationError(msg)
            elif not click.confirm(msg + ". Remove this tag so we can add it to the new snapshot?"):
                raise UserAbort()
            else:
                mixin.remove_tag_from_snapshot(existing_tag_md.hashval, tag)

    (md, manifest) = mixin.snapshot(tag, message)

    try:
        old_md = mixin.get_snapshot_metadata(md.hashval)  # type: Optional[SnapshotMetadata]
    except:
        old_md = None
    if old_md is not None:
        md = merge_snapshot_metadata(old_md, md, workspace.batch)

    mixin.save_snapshot_metadata_and_manifest(md, manifest)
    workspace.save("Completed snapshot %s" % md.hashval)

    if tag:
        click.echo(
            "Have successfully taken snapshot of workspace, tagged with '%s', hash is %s."
            % (tag, md.hashval)
        )
    else:
        click.echo("Have successfully taken snapshot of workspace, hash is %s." % md.hashval)
    return md.hashval



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
import click
from typing import cast


from dataworkspaces.errors import ConfigurationError, UserAbort
from dataworkspaces.workspace import Workspace, SnapshotWorkspaceMixin


def delete_snapshot_command(
    workspace: Workspace, tag_or_hash: str, no_include_resources: bool = True
) -> None:
    if not isinstance(workspace, SnapshotWorkspaceMixin):
        raise ConfigurationError("Workspace %s does not support snapshots." % workspace.name)
    mixin = cast(SnapshotWorkspaceMixin, workspace)
    md = mixin.get_snapshot_by_tag_or_hash(tag_or_hash)
    snapshot_name = (
        "%s (Tagged as: %s)" % (md.hashval[0:7], ", ".join(md.tags))
        if md.tags is not None
        else md.hashval
    )
    if not workspace.batch:
        if not click.confirm(
            "Should I delete snapshot %s? This is not reversible." % snapshot_name
        ):
            raise UserAbort()
    mixin.delete_snapshot(md.hashval, include_resources=not no_include_resources)
    workspace.save("Deleted snapshot %s" % snapshot_name)
    click.echo("Successfully deleted snapshot %s." % snapshot_name)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Resource for files living in a local directory 
"""
from errno import EEXIST
import os
from os.path import join, exists, isdir
from typing import List, Pattern, Tuple, Optional, Set, Union, cast
import json
import shutil

import click

from dataworkspaces.errors import ConfigurationError, PathError
from dataworkspaces.utils.subprocess_utils import call_subprocess
from dataworkspaces.utils.file_utils import does_subpath_exist, LocalPathType
from dataworkspaces.utils.git_utils import GIT_EXE_PATH, is_git_staging_dirty
from dataworkspaces.workspace import (
    Workspace,
    Resource,
    LocalStateResourceMixin,
    FileResourceMixin,
    SnapshotResourceMixin,
    SnapshotWorkspaceMixin,
    JSONDict,
    JSONList,
    ResourceFactory,
)
import dataworkspaces.resources.hashtree as hashtree
from dataworkspaces.utils.snapshot_utils import (
    move_current_files_local_fs,
    copy_current_files_local_fs,
)
import dataworkspaces.backends.git as git_backend
from dataworkspaces.utils.param_utils import StringType, BoolType


LOCAL_FILE = "file"


def _relative_rsrc_dir_for_git_workspace(role, name):
    return ".dataworkspace/" + LOCAL_FILE + "/" + role + "/" + name


class LocalFileResource(
    Resource, LocalStateResourceMixin, FileResourceMixin, SnapshotResourceMixin
):
    """Resource class for local files. Can also serve as the base class for other resource
    types (e.g. rclone)."""

    def __init__(
        self,
        resource_type: str,
        name: str,
        role: str,
        workspace: Workspace,
        global_local_path: str,
        my_local_path: Optional[str],
        compute_hash: Optional[bool] = None,
        export: Optional[bool] = None,
        imported: Optional[bool] = None,
        ignore: Optional[List[str]] = None,
    ):
        super().__init__(resource_type, name, role, workspace)
        self.param_defs.define(
            "global_local_path",
            default_value=None,
            optional=False,
            is_global=True,
            help="Location of files on local filesystem, as defined when the resource is created. "
            + "May be overridden locally via my_local_path.",
            ptype=StringType(),
        )
        self.global_local_path = self.param_defs.get(
            "global_local_path", global_local_path
        )  # type: str
        self.param_defs.define(
            "my_local_path",
            default_value=None,
            optional=True,
            is_global=False,
            help="Override of global_local_path, just for this instance of the workspace.",
            ptype=StringType(),
        )
        self.my_local_path = self.param_defs.get(
            "my_local_path", my_local_path
        )  # type: Optional[str]
        # the actual local path we'll use
        self.local_path = (
            self.my_local_path if self.my_local_path is not None else self.global_local_path
        )
        self.param_defs.define(
            "compute_hash",
            default_value=False,
            optional=False,
            is_global=True,
            help="If True, then compute the full hash of all files rather than using sizes.",
            ptype=BoolType(),
        )
        self.compute_hash = self.param_defs.get("compute_hash", compute_hash)  # type: bool
        self.param_defs.define(
            "export",
            default_value=False,
            optional=False,
            help="True if metadata for export should be added each snapshot",
            is_global=True,
            ptype=BoolType(),
            allow_missing=True,
        )
        self.export = self.param_defs.get("export", export)  # type: bool
        self.param_defs.define(
            "imported",
            default_value=False,
            optional=False,
            help="If True, then this resource has lineage imported from another workspace",
            is_global=True,
            ptype=BoolType(),
            allow_missing=True,
        )
        self.imported = self.param_defs.get("imported", imported)  # type: bool
        self.ignore = ignore if (ignore is not None) else []  # TODO: should this be a parameter?
        if isinstance(workspace, git_backend.Workspace):
            # if the workspace is a git repo, then we can store our
            # hash files there.
            self.rsrcdir = os.path.join(
                workspace.workspace_dir, _relative_rsrc_dir_for_git_workspace(self.role, self.name)
            )
        else:
            # If the workspace is not a git repo, we store the hash
            # files under the local path. Note that this is not going
            # to be replicated to other workspace instances, unless you
            # are using a shared file system (e.g. NFS).
            # TODO: add an API to the workspace for storing this arbitrary
            # data. It will also need changes to the hashtree file.
            self.rsrcdir = os.path.abspath(os.path.join(self.local_path, ".hashes"))
            self.ignore.append(".hashes")

    def get_local_path_if_any(self):
        return self.local_path

    def results_move_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ) -> None:
        move_current_files_local_fs(
            self.name,
            self.local_path,
            rel_dest_root,
            exclude_files,
            exclude_dirs_re,
            verbose=self.workspace.verbose,
        )

    def results_copy_current_files(
        self, rel_dest_root: str, exclude_files: Set[str], exclude_dirs_re: Pattern
    ) -> None:
        copy_current_files_local_fs(
            self.name,
            self.local_path,
            rel_dest_root,
            exclude_files,
            exclude_dirs_re,
            verbose=self.workspace.verbose,
        )

    def add_results_file(self, data: Union[JSONDict, JSONList], rel_dest_path: str) -> None:
        """save JSON results data to the specified path in the resource.
        """
        # assert self.role == ResourceRoles.RESULTS
        abs_dest_path = os.path.join(self.local_path, rel_dest_path)
        parent_dir = os.path.dirname(abs_dest_path)
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir)
        with open(abs_dest_path, "w") as f:
            json.dump(data, f, indent=2)

    def upload_file(self, local_path: str, rel_dest_path: str) -> None:
        """Copy a local file to the specified path in the
        resource. This may be a local copy or an upload, depending
        on the resource implmentation
        """
        abs_dest_path = os.path.join(self.local_path, rel_dest_path)
        parent_dir = os.path.dirname(abs_dest_path)
        if not exists(local_path):
            raise ConfigurationError("Source file %s does not exist." % local_path)
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir)
        shutil.copyfile(local_path, rel_dest_path)

    def does_subpath_exist(
        self, subpath: str, must_be_file: bool = False, must_be_directory: bool = False
    ) -> bool:
        return does_subpath_exist(self.local_path, subpath, must_be_file, must_be_directory)

    def delete_file(self, rel_path: str) -> None:
        os.remove(os.path.join(self.local_path, rel_path))

    def read_results_file(self, subpath: str) -> JSONDict:
        """Read and parse json results data from the specified path
        in the resource. If the path does not exist or is not a file
        throw a ConfigurationError.
        """
        path = os.path.join(self.local_path, subpath)
        if not os.path.isfile(path):
            raise ConfigurationError(
                "subpath %s does not exist or is not a file in resource %s" % (subpath, self.name)
            )
        with open(path, "r") as f:
            try:
                return json.load(f)
            except Exception as e:
                raise ConfigurationError(
                    "Parse error when reading %s in resource %s" % (subpath, self.name)
                ) from e

    def ls(self, rel_path:str) -> List[str]:
        if rel_path!='':
            path = os.path.join(self.local_path, rel_path)
        else:
            path = self.local_path
        if isdir(path):
            files = os.listdir(path)
            if rel_path!='':
                return [join(rel_path, file) for file in files]
            else:
                return files
        elif exists(path):
            return [rel_path]
        else:
            raise PathError(f"Path {rel_path} does not exist in resource {self.name}")

    def open(self, rel_path:str, mode:str):
        if rel_path=='':
            raise PathError("Need to specify a file to open, empty string is not valid here")
        path = os.path.join(self.local_path, rel_path)
        return open(path, mode)

    def get_local_params(self) -> JSONDict:
        return {"local_path": self.my_local_path}

    def pull_precheck(self) -> None:
        """Nothing to do, since we donot support sync.
        """
        pass

    def pull(self) -> None:
        """Nothing to do, since we donot support sync.
        """
        pass

    def push_precheck(self) -> None:
        """Nothing to do, since we donot support sync.
        """
        pass

    def push(self) -> None:
        """Nothing to do, since we donot support sync.
        """
        pass

    def snapshot_precheck(self) -> None:
        pass

    def snapshot(self) -> Tuple[Optional[str], Optional[str]]:
        if self.compute_hash:
            h = hashtree.generate_sha_signature(
                self.rsrcdir, self.local_path, ignore=self.ignore, verbose=self.workspace.verbose
            )
        else:
            h = hashtree.generate_size_signature(
                self.rsrcdir, self.local_path, ignore=self.ignore, verbose=self.workspace.verbose
            )
        assert os.path.exists(os.path.join(self.rsrcdir, h))
        if isinstance(self.workspace, git_backend.Workspace):
            workspace_path = self.workspace.get_workspace_local_path_if_any()
            assert workspace_path is not None
            if is_git_staging_dirty(
                workspace_path, subdir=_relative_rsrc_dir_for_git_workspace(self.role, self.name)
            ):
                call_subprocess(
                    [
                        GIT_EXE_PATH,
                        "commit",
                        "-m",
                        "Add snapshot hash files for resource %s" % self.name,
                    ],
                    cwd=workspace_path,
                    verbose=self.workspace.verbose,
                )
        return (h, None)

    def restore_precheck(self, hashval):
        # TODO: look at handling of restore - we probably want to do a compare and error out if
        # different. This would mean passing in both the compare and restore hashes.
        if self.compute_hash:
            rc = hashtree.check_sha_signature(
                hashval,
                self.rsrcdir,
                self.local_path,
                ignore=self.ignore,
                verbose=self.workspace.verbose,
            )
        else:
            rc = hashtree.check_size_signature(
                hashval,
                self.rsrcdir,
                self.local_path,
                ignore=self.ignore,
                verbose=self.workspace.verbose,
            )
        if not rc:
            print("ERROR IN RESTORE")
            raise ConfigurationError("Local file structure not compatible with saved hash")

    def restore(self, hashval):
        pass  # local files: do nothing to restore

    def delete_snapshot(
        self, workspace_snapshot_hash: str, resource_restore_hash: str, relative_path: str
    ) -> None:
        snapshot_dir_path = join(self.local_path, relative_path)
        if isdir(snapshot_dir_path):
            if self.workspace.verbose:
                print(
                    "Deleting snapshot directory %s from resource %s" % (relative_path, self.name)
                )
            shutil.rmtree(snapshot_dir_path)

    def validate_subpath_exists(self, subpath: str) -> None:
        super().validate_subpath_exists(subpath)

    def __str__(self):
        return "Local directory %s in role '%s'" % (self.local_path, self.role)


def setup_path_for_hashes(role: str, name: str, workspace: Workspace, local_path: str):
    """When creating the resource, make sure we have a path for the hashes.
    Subclasses of LocalFilesResource will need to call this in their factory
    functions."""
    workspace_path = workspace.get_workspace_local_path_if_any()
    if isinstance(workspace, git_backend.Workspace):
        assert workspace_path is not None
        hash_path = join(workspace_path, _relative_rsrc_dir_for_git_workspace(role, name))
        try:
            os.makedirs(hash_path)
            with open(os.path.join(hash_path, "dummy.txt"), "w") as f:
                f.write("Placeholder to ensure directory is added to git\n")
            call_subprocess(
                [
                    GIT_EXE_PATH,
                    "add",
                    join(_relative_rsrc_dir_for_git_workspace(role, name), "dummy.txt"),
                ],
                cwd=workspace_path,
            )
            call_subprocess(
                [GIT_EXE_PATH, "commit", "-m", "Adding resource %s" % name], cwd=workspace_path
            )
        except OSError as exc:
            if exc.errno == EEXIST and os.path.isdir(hash_path):
                pass
            else:
                raise
    else:
        non_git_hashes = join(local_path, ".hashes")
        if not exists(non_git_hashes):
            os.mkdir(non_git_hashes)


class LocalFileFactory(ResourceFactory):
    def from_command_line(self, role, name, workspace, local_path, compute_hash, export, imported):
        """Instantiate a resource object from the add command's arguments"""
        if not os.path.isdir(local_path):
            raise ConfigurationError(local_path + " does not exist")
        if not os.access(local_path, os.R_OK):
            raise ConfigurationError(local_path + " does not have read permission")
        setup_path_for_hashes(role, name, workspace, local_path)
        if imported:
            lineage_path = join(local_path, "lineage.json")
            if not exists(lineage_path):
                raise ConfigurationError(
                    "--imported was specified, but missing exported lineage file %s" % lineage_path
                )
            if (
                not isinstance(workspace, SnapshotWorkspaceMixin)
                or not workspace.supports_lineage()
            ):
                raise ConfigurationError(
                    "--imported was specified, but this workspace does not support lineage"
                )
            with open(lineage_path, "r") as f:
                lineage_data = json.load(f)
            if lineage_data["resource_name"] != name:
                raise ConfigurationError(
                    "Resource name in imported lineage '%s' does not match '%s'"
                    % (lineage_data["resource_name"], name)
                )
            cast(SnapshotWorkspaceMixin, workspace).get_lineage_store().import_lineage_file(
                name, lineage_data["lineages"]
            )

        return LocalFileResource(
            LOCAL_FILE,
            name,
            role,
            workspace,
            local_path,
            my_local_path=None,
            compute_hash=compute_hash,
            export=export,
            imported=imported,
        )

    def from_json(
        self, params: JSONDict, local_params: JSONDict, workspace: Workspace
    ) -> LocalFileResource:
        """Instantiate a resource object from saved params and local params"""
        return LocalFileResource(
            LOCAL_FILE,
            params["name"],
            params["role"],
            workspace,
            # for backward compatibility, we also check for "local_path"
            global_local_path=params["global_local_path"]
            if "global_local_path" in params
            else params["local_path"],
            my_local_path=local_params["my_local_path"]
            if "my_local_path" in local_params
            else (local_params["local_path"] if "local_path" in local_params else None),
            compute_hash=params["compute_hash"],
            export=params.get("export", None),
            imported=params.get("imported", None),
        )

    def has_local_state(self) -> bool:
        return True

    def clone(self, params: JSONDict, workspace: Workspace) -> LocalStateResourceMixin:
        """Instantiate a resource that was created remotely. We need to verify that
        the local copy of the data exists -- we are not responsible for making certain
        it is in th correct place.
        """
        name = params["name"]
        # check local_path, too for backward compatibility
        global_local_path = (
            params["global_local_path"] if "global_local_path" in params else params["local_path"]
        )  # type: str
        local_params = {}  # type: JSONDict
        if exists(global_local_path):
            local_path = global_local_path
        else:
            if not workspace.batch:
                local_path = cast(
                    str,
                    click.prompt(
                        "Local files resource '%s' was located at '%s' on the original system. W\here is it located on this system?"
                        % (name, global_local_path),
                        type=LocalPathType(exists=True),
                    ),
                )
                local_params["my_local_path"] = local_path
            else:
                raise ConfigurationError(
                    "Local files resource %s is missing from %s." % (name, global_local_path)
                )
        if not isinstance(workspace, git_backend.Workspace):
            non_git_hashes = join(local_path, ".hashes")
            if not exists(non_git_hashes):
                os.mkdir(non_git_hashes)
        return self.from_json(params, local_params, workspace)

    def suggest_name(self, workspace, role, local_path, compute_hash, export, imported):
        return os.path.basename(local_path)



"""
This module (``dataworkspaces.kits.scikit_learn``)
provides integration with the `scikit-learn <https://scikit-learn.org>`_
framework. The main class provided here is :class:`~LineagePredictor`,
which wraps any class following sklearn's predictor protocol. It captures
inputs, model parameters and results. This module also provides
:class:`~Metrics` and its subclasses, which
support the computation of common metrics and the writing of them
to a results file. Finally, there is
:func:`~train_and_predict_with_cv`, which
runs a common sklearn classification workflow, including grid search.
"""

from typing import Optional, Union, Dict, List, Any, cast
from abc import ABCMeta, abstractmethod
from sklearn.base import ClassifierMixin # type: ignore
from sklearn import metrics # type: ignore
from sklearn.model_selection import GridSearchCV, train_test_split # type: ignore
from sklearn.utils import Bunch # type: ignore
import sklearn.utils.metaestimators # type: ignore
import sys
import numpy as np # type: ignore
import os
from os.path import join, abspath, expanduser, exists, isabs
from tempfile import NamedTemporaryFile


from dataworkspaces.errors import ConfigurationError
from dataworkspaces.lineage import LineageBuilder
from dataworkspaces.workspace import (
    find_and_load_workspace,
    LocalStateResourceMixin,
    FileResourceMixin,
)
from dataworkspaces.utils.lineage_utils import ResourceRef
from dataworkspaces.kits.wrapper_utils import _DwsModelState, _add_to_hash

from .jupyter import is_notebook, get_step_name_for_notebook, get_notebook_directory

try:
    import joblib # type: ignore
except ImportError as e:
    raise ConfigurationError('Please install the joblib package (via "pip install joblib")') from e


def _load_dataset_file(dataset_path, filename):
    filepath = join(dataset_path, filename)
    if filename.endswith(".txt") or filename.endswith(".rst"):
        with open(filepath, "r") as f:
            return f.read()
    elif filename.endswith(".csv") or filename.endswith(".csv.gz") or filename.endswith(".csv.bz2"):
        try:
            return np.loadtxt(filepath, delimiter=",")
        except ValueError:
            # try with pandas
            import pandas # type: ignore

            df = pandas.read_csv(filepath)
            if len(df.values.shape) == 2 and df.values.shape[1] == 1:  # this is just a list
                return df.values.reshape(df.values.shape[0])
            else:
                return df.values
    elif filename.endswith(".npy"):
        return np.load(filepath)


def load_dataset_from_resource(
    resource_name: str, subpath: Optional[str] = None, workspace_dir: Optional[str] = None
) -> Bunch:
    """
    Load a datset (data and targets) from the specified resource, and returns an
    sklearn-style Bunch (a dictionary-like object). The bunch will include at least
    three attributes:

    * ``data`` - a NumPy array of shape number_samples * number_features
    * ``target`` - a NumPy array of length number_samples
    * ``resource`` - a :class:`~dataworkspaces.workspace.ResourceRef` that provides the resource name and
      subpath (if any) for the data

    Some other attributes that may also be present, depending on the data set:

    * ``DESCR`` - text containing a full description of the data set (for humans)
    * ``feature_names`` - an array of length number_features containing the name
      of each feature.
    * ``target_names`` - an array containing the name of each target class

    Data sets may define their own attributes as well (see below).

    **Parameters**

    resource_name
        The name of the resource containing the dataset.

    subpath
        Optional subpath within the resource where this specific dataset is located.
        If not specified, the root of the resource is used.

    workspace_dir
       The root directory of your workspace in the local file system. Usually,
       this can be left unspecified and inferred by DWS, which will search up
       from the current working directory.

    **Creating a Dataset**

    To create a dataset in your resource that is suitable for importing by this function,
    you simply need to create a file for each attribute you want in the bunch and place
    all these files in the same directory within your resource.
    The names of the files should be ``ATTRIBUTE.extn`` where ``ATTRIBUTE`` is the
    attribute name (e.g. ``data`` or ``DESCR``) and ``.extn`` is a file extension
    indicating the format. Supported file extensions are:

    * ``.txt`` or ``.rst`` - text files
    * ``.csv`` - csv files. These are read in using ``numpy.loadtxt()``. If this
      fails because the csv does not contain all numeric data, pandas is used to read
      in the file. It is then converted back to a numpy array.
    * ``.csv.gz`` or ``.csv.bz2`` - these are compressed csv files which are treated
      the same was as csv files (numpy and pandas will automatically uncompress before parsing).
    * ``.npy`` - this a a file containing a serialized NumPy array saved via ``numpy.save()``.
      It is loaded using ``numpy.load()``.
    """

    workspace = find_and_load_workspace(True, False, workspace_dir)
    workspace.validate_resource_name(resource_name, subpath)
    dataset_name = (
        "Resource " + resource_name + " subpath " + subpath
        if subpath is not None
        else "Resource " + resource_name
    )
    r = workspace.get_resource(resource_name)
    if not isinstance(r, LocalStateResourceMixin) or (r.get_local_path_if_any() is None):
        # TODO: Support a data access api
        raise ConfigurationError(
            "Unable to instantiate a data set for resource '%s': currently not supported for non-local resources"
            % resource_name
        )
    local_path = r.get_local_path_if_any()
    assert local_path is not None
    dataset_path = join(local_path, subpath) if subpath is not None else local_path
    result = {}  # this will be the args to the result Bunch
    # First load data and target files, which are required
    data_file = join(dataset_path, "data.csv")
    if exists(data_file):
        pass
    elif exists(data_file + ".gz"):
        data_file += ".gz"
    elif exists(data_file + ".bz2"):
        data_file += ".bz2"
    else:
        raise ConfigurationError(
            "Did not find data file for %s at '%s'" % (dataset_name, data_file)
        )
    result["data"] = np.loadtxt(data_file, delimiter=",")
    target_file = join(dataset_path, "target.csv")
    if exists(target_file):
        pass
    elif exists(target_file + ".gz"):
        target_file += ".gz"
    elif exists(target_file + ".bz2"):
        target_file += ".bz2"
    else:
        raise ConfigurationError(
            "Did not find target file for %s at '%s'" % (dataset_name, target_file)
        )
    result["target"] = np.loadtxt(target_file, delimiter=",")
    if result["data"].shape[0] != result["target"].shape[0]:
        raise ConfigurationError(
            "Data matrix at '%s' has %d rows, but target at '%s' has %d rows"
            % (data_file, result["data"].shape[0], target_file, result["target"].shape[0])
        )
    result["resource"] = ResourceRef(resource_name, subpath)
    # check for and load any other attributes
    for fname in os.listdir(dataset_path):
        if fname.endswith(".txt"):
            result[fname[:-4]] = _load_dataset_file(dataset_path, fname)
        elif fname.endswith(".rst"):
            result[fname[:-4]] = _load_dataset_file(dataset_path, fname)
        elif fname.endswith(".csv"):
            result[fname[:-4]] = _load_dataset_file(dataset_path, fname)
        elif fname.endswith(".csv.gz"):
            result[fname[:-7]] = _load_dataset_file(dataset_path, fname)
        elif fname.endswith(".csv.bz2"):
            result[fname[:-8]] = _load_dataset_file(dataset_path, fname)
        elif fname.endswith(".npy"):
            result[fname[:-4]] = _load_dataset_file(dataset_path, fname)
    return Bunch(**result)


class Metrics(metaclass=ABCMeta):
    """Metrics and its subclasses are convenience classes
    for sklearn metrics. The subclasses
    of Matrics are used by :func:`~train_and_predict_with_cv`
    in printing a metrics report and generating the metrics
    json file.
    """

    def __init__(self, expected, predicted, sample_weight=None):
        self.expected = expected
        self.predicted = predicted
        self.sample_weight = sample_weight

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def score(self) -> float:
        """Given the expected and predicted values, compute the metric
        for this type of predictor, as needed for the predictor's score()
        method. This is used in the wrapped classes to avoid multiple
        calls to predict()."""
        pass

    @abstractmethod
    def print_metrics(self, file=sys.stdout) -> None:
        """Print the metrics to a file
        """
        pass


class BinaryClassificationMetrics(Metrics):
    """Given an array of expected (target) values
    and the actual predicted values from a classifier,
    compute metrics that make sense for a binary
    classifier, including accuracy, precision, recall, roc auc,
    and f1 score.
    """

    def __init__(self, expected, predicted, sample_weight=None):
        super().__init__(expected, predicted, sample_weight)
        self.accuracy = metrics.accuracy_score(expected, predicted, sample_weight=sample_weight)
        self.precision = metrics.precision_score(expected, predicted, sample_weight=sample_weight)
        self.recall = metrics.recall_score(expected, predicted, sample_weight=sample_weight)
        self.roc_auc = metrics.roc_auc_score(expected, predicted, sample_weight=sample_weight)
        self.f1_score = metrics.f1_score(expected, predicted, sample_weight=sample_weight)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "accuracy": self.accuracy,
            "precision": self.precision,
            "recall": self.recall,
            "roc_auc_score": self.roc_auc,
            "f1_score": self.f1_score,
        }

    def score(self) -> float:
        """Metric for binary classification is accuracy
        """
        return self.accuracy

    def print_metrics(self, file=sys.stdout) -> None:
        for k, v in self.to_dict().items():
            print("%13s: %.02f" % (k, cast(Union[int, float], v)), file=file)


class MulticlassClassificationMetrics(Metrics):
    """Given an array of expected (target) values
    and the actual predicted values from a classifier,
    compute metrics that make sense for a multi-class
    classifier, including accuracy and sklearn's
    "classification report" showing per-class metrics.
    """

    def __init__(self, expected, predicted, sample_weight=None):
        super().__init__(expected, predicted, sample_weight)
        self.accuracy = metrics.accuracy_score(expected, predicted, sample_weight=sample_weight)
        self.classification_report = metrics.classification_report(
            expected, predicted, sample_weight=sample_weight, output_dict=True
        )

    def score(self) -> float:
        """Metric for multiclass classification is accuracy
        """
        return self.accuracy

    def to_dict(self):
        return {"accuracy": self.accuracy, "classification_report": self.classification_report}

    def print_metrics(self, file=sys.stdout):
        print("accuracy: %.02f" % self.accuracy, file=file)
        print("classification report:", file=file)
        print(metrics.classification_report(self.expected, self.predicted))


class RegressionMetrics(Metrics):
    """For regression, we capture the r-squared score and
    the mean squared error.
    """

    def __init__(self, expected, predicted, sample_weight=None):
        super().__init__(expected, predicted, sample_weight)
        self.r2_score = metrics.r2_score(self.expected, self.predicted, sample_weight=sample_weight)
        self.mean_squared_error = metrics.mean_squared_error(
            self.expected, self.predicted, sample_weight=sample_weight
        )

    def score(self) -> float:
        """Metric for regression is r2_score
        """
        return self.r2_score

    def to_dict(self):
        return {"r2_score": self.r2_score, "mean_squared_error": self.mean_squared_error}

    def print_metrics(self, file=sys.stdout):
        print("r2_score: %.02f" % self.r2_score, file=file)
        print("mean_squared_error: %.02f" % self.mean_squared_error, file=file)


_METRICS = {
    "binary_classification": BinaryClassificationMetrics,
    "multiclass_classification": MulticlassClassificationMetrics,
    "regression": RegressionMetrics,
}


class LineagePredictor(sklearn.utils.metaestimators._BaseComposition):
    """This is a wrapper for adding lineage to any predictor in sklearn.
    To use it, instantiate the predictor (for classification or regression)
    and then create a new instance of :class:`~LineagePredictor`.

    The initializer finds the associated workspace and initializes a
    :class:`~dataworkspaces.lineage.Lineage` instance. The input_resource
    is recorded in this lineage. Other methods call the underlying wrapped
    predictor's methods, with additional functionality as needed (see below).

    **Parameters**

    predictor
        Any sklearn predictor instance. It must have ``fit`` and ``predict``
        methods.

    metrics
        Either a string naming a metrics type or a subclass of :class:`~Metrics`.
        If a string, it should be one of: binary_classification,
        multiclass_classification, or regression.

    input_resource
        Resource providing the input data to this model. May be
        specified by name, by a local file path, or via a
        :class:`~dataworkspaces.workspace.ResourceRef`.

    resource_resource
        (optional) Resource where the results are to be stored.
        May be specified by name, by a local file path, or via a
        :class:`!ResourceRef`.
        If not specified, will try to infer from the workspace.

    model_save_file
        (optional) Name of file to store a (joblib-formmatted)
        serialization of the trained model upon completion of the ``fit()``
        method. This should be a relative path, as it is stored under
        the results resource. If model_save_file is not specified,
        no model is saved.

    workspace_dir
        (optional) Directory specifying the workspace. Usually can be
        inferred from the current directory.

    verbose
        If True, print a lot of detailed information about the execution
        of Data Workspaces.

    **Example**

    Here is an example useage of the wrapper, taken from the
    :ref:`Quick Start <quickstart>`::

      from sklearn.svm import SVC
      from sklearn.model_selection import train_test_split
      from dataworkspaces.kits.scikit_learn import load_dataset_from_resource
      from dataworkspaces.kits.scikit_learn import LineagePredictor

      dataset = load_dataset_from_resource('sklearn-digits-dataset')
      X_train, X_test, y_train, y_test = train_test_split(
          dataset.data, dataset.target, test_size=0.5, shuffle=False)
      classifier = LineagePredictor(SVC(gamma=0.001),
                                    metrics='multiclass_classification',
                                    input_resource=dataset.resource,
                                    model_save_file='digits.joblib')

      classifier.fit(X_train, y_train)
      score = classifier.score(X_test, y_test)

    **Methods**

    """

    _dws_model_wrap = True

    def __init__(
        self,
        predictor,
        metrics: Union[str, type],
        input_resource: Union[str, ResourceRef],
        results_resource: Optional[Union[str, ResourceRef]] = None,
        model_save_file: Optional[str] = None,
        workspace_dir: Optional[str] = None,
        verbose: bool = False,
    ):
        if hasattr(predictor, "_dws_model_wrap") and predictor._dws_model_wrap is True:  # type: ignore
            print("dws>> %s is already wrapped" % repr(predictor))
            return predictor  # already wrapped
        self.predictor = predictor
        assert metrics in _METRICS.keys() or (
            isinstance(metrics, type) and issubclass(metrics, Metrics)
        ), (
            "%s is not a subclass of Metrics and not one of %s"
            % (repr(metrics), ", ".join([repr(s) for s in _METRICS.keys()]))
        )
        self.metrics = metrics
        self.input_resource = input_resource
        self.results_resource = results_resource
        self.model_save_file = model_save_file
        if model_save_file is not None:
            assert not isabs(model_save_file), "Model save file should not be an absolute path"
        self.workspace_dir = workspace_dir
        self.metrics = metrics
        self.verbose = verbose
        self.score_has_been_run = False
        self._init_dws_state()

    def _init_dws_state(self):
        workspace = find_and_load_workspace(
            batch=True, verbose=self.verbose, uri_or_local_path=self.workspace_dir
        )
        self._dws_state = _DwsModelState(workspace, self.input_resource, self.results_resource)

    def _save_model(self):
        assert self.model_save_file
        if not self.model_save_file.endswith(".joblib"):
            model_save_file = self.model_save_file + ".joblib"
        else:
            model_save_file = self.model_save_file
        tempname = None
        try:
            with NamedTemporaryFile(delete=False, suffix=".joblib") as f:
                tempname = f.name
            joblib.dump(self, tempname)
            resource = self._dws_state.workspace.get_resource(self._dws_state.results_ref.name)
            if self._dws_state.results_ref.subpath is not None:
                target_name = join(self._dws_state.results_ref.subpath, model_save_file)
            else:
                target_name = model_save_file
            cast(FileResourceMixin, resource).upload_file(tempname, target_name)
        finally:
            if (tempname is not None) and exists(tempname):
                os.remove(tempname)
            if self.verbose:
                print("dws> saved model file to %s:%s" % (resource.name, target_name))

    def __getstate__(self):
        state = super().__getstate__()
        if "_dws_state" in state:
            del state["_dws_state"]
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self._init_dws_state()

    def set_params(self, **params):
        """"""
        super().set_params(**params)
        self._init_dws_state()
        return self

    def fit(self, X, y, *args, **kwargs):
        """The underlying fit() method of a predictor trains the predictio based
        on the input data (X) and labels (y).

        If the input resource is an api resource, the wrapper captures the hash of
        the inputs.
        If ``model_save_file`` was specified, it also saves the trained model."""
        api_resource = self._dws_state.find_input_resources_and_return_if_api(X, y)
        if api_resource is not None:
            api_resource.init_hash_state()
            hash_state = api_resource.get_hash_state()
            _add_to_hash(X, hash_state)
            _add_to_hash(y, hash_state)
            api_resource.save_current_hash()  # in case we evaluate in a separate process
        result = self.predictor.fit(X, y, *args, **kwargs)
        if self.model_save_file is not None:
            self._save_model()
        return result

    def score(self, X, y, sample_weight=None):
        """This method make predictions from a trained model and scores them
        according to the metrics specified when instantiated the wrapper.

        If the input resource is an api resource, the wrapper captures its hash.
        The wapper runs the wrapped predictor's :meth:`~predict` method to
        generate predictions. A `metrics` object is instantiated to compute the metrics
        for the predictions and a ``results.json`` file is written to the
        results resource. The lineage data is saved and finally the score
        is computed from the predictions and returned to the caller."""
        if self.score_has_been_run:
            # This might be from a saved model, so we reset the
            # execution time, etc.
            self._dws_state.reset_lineage()
        for (param, value) in self.predictor.get_params(deep=True).items():
            self._dws_state.lineage.add_param(param, value)
        api_resource = self._dws_state.find_input_resources_and_return_if_api(X, y)
        if api_resource is not None:
            api_resource.dup_hash_state()
            hash_state = api_resource.get_hash_state()
            _add_to_hash(X, hash_state)
            if y is not None:
                _add_to_hash(y, hash_state)
            api_resource.save_current_hash()
            api_resource.pop_hash_state()
        predictions = self.predictor.predict(X)
        if isinstance(self.metrics, str):
            metrics_inst = _METRICS[self.metrics](y, predictions, sample_weight=sample_weight)  # type: ignore
        else:
            metrics_inst = self.metrics(y, predictions, sample_weight=sample_weight)
        self._dws_state.write_metrics_and_complete(metrics_inst.to_dict())
        self.score_has_been_run = True
        return metrics_inst.score()

    def predict(self, X):
        """The underlying :meth:`~predict` method is called directly,
        without affecting the lineage."""
        return self.predictor.predict(X)


def train_and_predict_with_cv(
    classifier_class: ClassifierMixin,
    param_grid: Union[Dict[str, List[Any]], List[Dict[str, List[Any]]]],
    dataset: Bunch,
    results_dir: str,
    test_size: float = 0.2,
    folds: int = 5,
    cv_scoring: str = "accuracy",
    model_name: Optional[str] = None,
    random_state: Optional[int] = None,
    run_description: Optional[str] = None,
) -> None:
    """NOTE: This function is under consideration for DEPRECATION and
    may be removed from a future version.

    This function implements a common workflow for sklearn classifiers:

    1. Splits the data into training set and a final validation set.
    2. Runs a grid search cross validation to find the best combination
       of hyperparameters for the classifier on the training data.
    3. Trains the model on the training data using the best hyperparameter
       values.
    4. Predicts the classes of the validation test data set and computes
       common metrics comparing the training and testing data.
    5. Writes the metrics to a results file at RESULTS_DIR/results.json.
    6. If the ``model_name`` parameter was specified, retrain the classifier
       on all the data and save the (pickled) model to the results directory.
    7. Write out the lineage data for this experiment.

    **Parameters**

    classifier_class
        An sklearn classifier or a class implementing the same interface
    param_grid
        As described in the documentation for
        `GridSearchCV <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html#sklearn.model_selection.GridSearchCV>`_,
        a dictionary with parameters names as keys and lists of parameter settings to
        try as values, or a list of such dictionaries. The various combinations
        of these parameters will be searched to find the best classifiation results
        on the training data.
    dataset
        A sklean Bunch object with members for data, target, and resource. This can
        be loaded by calling :func:`~load_dataset_from_resource`
    results_dir
        The directory on the local filesystem to which the results should be
        written.
    test_size
        The fraction of the input data samples to be held back for the final
        validation. Defaults to 0.2 (20%).
    folds
        Number of cross validation folds. Defaults to 5.
    cv_scoring
        Name of scoring algorithm to use in evaluating the hyperparameter
        combinations in cross validation. Defaults to 'accuracy'. See
        `here <https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_
        for details.
    model_name
        If specified, retrain the model using the entire data set (train + test)
        and the best parameters found during cross validation. Pickle and save
        this model to the file RESULTS_DIR/MODEL_NAME.pkl. If the model name is
        not specified, skip this step.
    random_state
        Optional integer to be used as a random seed.
    run_description
        Optional text describing this particular run. This is saved in the results
        file and the lineage.

    **Example**

    Here is an example (taken from the :ref:`Quick Start <quickstart>`)::

        import numpy as np
        from os.path import join
        from sklearn.svm import SVC
        from dataworkspaces.kits.scikit_learn import load_dataset_from_resource,\
                                                     train_and_predict_with_cv
        
        RESULTS_DIR='../results'
        
        dataset = load_dataset_from_resource('sklearn-digits-dataset')
        train_and_predict_with_cv(SVC, {'gamma':[0.01, 0.001, 0.0001]}, dataset,
                                  RESULTS_DIR, random_state=42)

    This trains a Support Vector Classifier with three different values of gamma
    (0.01, 0.001, and 0.0001) and then evaluates the trained classifier on the
    holdout data. The results are writen to ``../results/results.json``.
    """
    X_train, X_test, y_train, y_test = train_test_split(
        dataset.data, dataset.target, test_size=test_size, random_state=random_state
    )
    # find the best combination of hyperparameters
    search = GridSearchCV(classifier_class(), param_grid=param_grid, scoring=cv_scoring, cv=folds)
    search.fit(X_train, y_train)
    best_params = search.best_params_
    print("Best params were: %s" % repr(best_params))

    lineage_params = {
        "classifier": classifier_class.__name__,
        "test_size": test_size,
        "cv_params": param_grid,
        "cv_scoring": cv_scoring,
        "random_state": random_state,
    }
    for (k, v) in best_params.items():
        lineage_params[k] = v

    lb = (
        LineageBuilder()
        .with_parameters(lineage_params)
        .as_results_step(results_dir, run_description)
        .with_input_ref(dataset.resource)
    )
    if is_notebook():
        lb = lb.with_code_path(get_notebook_directory())
        step_name = get_step_name_for_notebook()
        if step_name is not None:
            lb = lb.with_step_name(step_name)  # not always able to determine this
    else:
        lb = lb.as_script_step()

    with lb.eval() as lineage:
        # Instantiate a classifier with the best parameters and train
        classifier = classifier_class(**best_params)
        classifier.fit(X_train, y_train)

        # Now predict the value of the digit on the test set
        predicted = classifier.predict(X_test)
        m = (
            MulticlassClassificationMetrics(y_test, predicted)
            if len(np.unique(dataset.target)) > 2
            else BinaryClassificationMetrics(y_test, predicted)
        )  # type: Metrics
        m.print_metrics()
        lineage.write_results(m.to_dict())

        if model_name is not None:
            classifier = classifier_class(**best_params)
            classifier.fit(dataset.data, dataset.target)
            model_file = join(abspath(expanduser(results_dir)), model_name + ".pkl")
            joblib.dump(classifier, model_file)
            print("Wrote trained model to %s" % model_file)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
import click
from typing import Optional, cast, Dict, List, Any

assert Dict and List and Any  # for pyflakes
from collections import Counter

from dataworkspaces.workspace import RESOURCE_ROLE_CHOICES, Workspace, SnapshotWorkspaceMixin
from dataworkspaces.utils.print_utils import print_columns, ColSpec


METRIC_NAME_WIDTH = 18
METRIC_VAL_WIDTH = 12

NUM_METRICS = 2


def print_snapshot_history(
    workspace: SnapshotWorkspaceMixin, reverse: bool = True, max_count: Optional[int] = None
):
    history = workspace.list_snapshots(reverse, max_count)
    # find the most common metrics
    mcounter = Counter()  # type: Counter
    for md in history:
        if md.metrics is not None:
            mcounter.update(md.metrics.keys())
    metric_names = [m for (m, cnt) in mcounter.most_common(NUM_METRICS)]
    spec = {
        "Hash": ColSpec(width=8),
        "Tags": ColSpec(width=20),
        "Created": ColSpec(width=19),
        "Message": ColSpec(width=30),
    }
    hashes = []  # type: List[str]
    tags = []  # type: List[str]
    created = []  # type: List[str]
    metrics = {n: [] for n in metric_names}  # type: Dict[str,List[Any]]
    messages = []  # type: List[str]

    returned = 0
    for md in history:
        hashes.append(md.hashval[0:7])
        tags.append(", ".join(md.tags))
        created.append(md.timestamp[0:-7])
        messages.append(md.message)
        for m in metric_names:
            metrics[m].append(md.metrics[m] if md.metrics is not None and m in md.metrics else None)
        returned += 1
    columns = {"Hash": hashes, "Tags": tags, "Created": created}
    for m in metric_names:
        columns[m] = metrics[m]
        spec[m] = ColSpec(width=25, truncate=True)
    columns["Message"] = messages
    click.echo("\n")
    print_columns(columns, null_value="", spec=spec, paginate=False, title="History of snapshots")
    if max_count is not None and returned == max_count:
        click.echo("Showing first %d snapshots" % max_count)
    else:
        click.echo("%d snapshots total" % returned)


def print_resource_status(workspace: Workspace):
    names_by_role = {role: [] for role in RESOURCE_ROLE_CHOICES}  # type:Dict[str,List[str]]
    resource_names = []
    roles = []
    types = []
    params = []
    missing_roles = []
    # we are going to order resources by role
    for rname in workspace.get_resource_names():
        role = workspace.get_resource_role(rname)
        names_by_role[role].append(rname)
    for role in RESOURCE_ROLE_CHOICES:
        if len(names_by_role[role]) > 0:
            for rname in names_by_role[role]:
                resource_names.append(rname)
                roles.append(role)
                types.append(workspace.get_resource_type(rname))
                params.append(
                    ",\n".join(
                        [
                            "%s=%s" % (pname, pval)
                            for (pname, pval) in workspace._get_resource_params(rname).items()
                            if pname not in ("resource_type", "name", "role")
                        ]
                    )
                )
        else:
            missing_roles.append(role)
    print_columns(
        {"Resource": resource_names, "Role": roles, "Type": types, "Parameters": params},
        # spec={'Parameters':ColSpec(width=40)},
        null_value="",
        title="Resources for workspace: %s" % workspace.name,
        paginate=False,
    )
    if len(missing_roles) > 0:
        click.echo("No resources for the following roles: %s." % ", ".join(missing_roles))


def status_command(workspace: Workspace, history: bool, limit: Optional[int] = None):
    print("Status for workspace: %s" % workspace.name)
    print_resource_status(workspace)
    if history:
        if not isinstance(workspace, SnapshotWorkspaceMixin):
            click.echo(
                "Workspace %s cannot perform snapshots, ignoring --history option" % workspace.name,
                err=True,
            )
        else:
            print_snapshot_history(
                cast(SnapshotWorkspaceMixin, workspace), reverse=True, max_count=limit
            )



"""
Utilities for printing and formatting
"""
from typing import Dict, Any, NamedTuple, List, Optional
import click


class ColSpec(NamedTuple):
    precision: Optional[int] = None
    width: Optional[int] = None  # if spec, specifies exact width. Does not include
    # one spec padding each side
    truncate: bool = False  # If True and width is specified, truncate column
    # rather than wrapping it.
    alignment: str = "auto"


def _truncate(s, width):
    truncated = False
    if "\n" in s:
        s = s[0, s.index("\n")]  # tuncate at the first newline
        truncated = True
    if len(s) > (width - 2):
        s = s[0 : width - 2]
        truncated = True
    if truncated:
        s += ".."
    return s


def pad_left(s, width, truncate=False):
    if truncate:
        s = _truncate(s, width)
    if "\n" in s:
        return "\n".join([pad_left(fragment, width) for fragment in s.split("\n")])
    if len(s) == width:
        return s
    if len(s) < width:
        return (" " * (width - len(s))) + s
    else:
        wrapped = []
        while len(s) >= width:
            wrapped.append(s[0:width])
            s = s[width:]
        if len(s) > 0:
            wrapped.append(pad_left(s, width))
        return "\n".join(wrapped)


def pad_right(s, width, truncate=False):
    if truncate:
        s = _truncate(s, width)
    if "\n" in s:
        return "\n".join([pad_right(fragment, width) for fragment in s.split("\n")])
    if len(s) == width:
        return s
    if len(s) < width:
        return s + (" " * (width - len(s)))
    else:
        wrapped = []
        while len(s) >= width:
            wrapped.append(s[0:width])
            s = s[width:]
        if len(s) > 0:
            wrapped.append(pad_right(s, width))
        return "\n".join(wrapped)


class FormattedColumns(NamedTuple):
    nitems: int
    headers: List[str]
    columns: Dict[str, List[str]]
    widths: List[int]


def format_columns(
    columns: Dict[str, Any], precision=-1, null_value: str = "None", spec: Dict[str, ColSpec] = {}
) -> FormattedColumns:
    """Format and pad the individual columns.

    A precision of -1 means to adjust based on whether absolute value is less than one.
    """
    str_cols = {}  # type: Dict[str,List[str]]
    # start an iteration to get the length of the first column
    nitems = 0
    for values in columns.values():
        nitems = len(values)
        break
    headers = []
    widths = []
    # now go through all columns and compute string representations
    for (col, values) in columns.items():
        nitems_for_col = len(values)
        assert nitems_for_col == nitems
        cspec = spec.get(col, None)
        if cspec is not None and cspec.precision is not None:
            col_precision = cspec.precision
        else:
            col_precision = precision
        max_value_width = 0
        strvalues = []
        all_numeric = True
        truncate = False
        # find the maxium width, determine if numeric, and do rounding
        for v in values:
            if isinstance(v, float):
                if col_precision != -1:
                    v = round(v, col_precision)
                elif v > -1.0 and v < 1.0:
                    v = round(v, 3)
                else:
                    v = round(v, 1)
            if not isinstance(v, (int, float)) and v is not None:
                all_numeric = False
            if v is not None and v != "":
                s = str(v)
            else:
                s = null_value
            if "\n" in s:
                # special case if there are line breaks
                fragments = s.split("\n")
                len_s = max([len(fragment) for fragment in fragments])
            else:
                len_s = len(s)
            if len_s > max_value_width:
                max_value_width = len_s
            strvalues.append(s)
        width_needed = max(len(col), max_value_width)
        if cspec is not None and cspec.width is not None:
            if width_needed >= cspec.width:
                colwidth = cspec.width
                if cspec.truncate:
                    assert cspec.width > 3, "Column too short to truncate"
                    truncate = True
            else:
                # if possible, we make the column narrower than
                # specified.
                colwidth = width_needed
        else:
            colwidth = width_needed
        headers.append(pad_right(col, colwidth))  # headers are always left aligned
        widths.append(colwidth)
        if cspec is not None and cspec.alignment != "auto":
            if cspec.alignment == "right":
                pad_fn = pad_left
            else:
                pad_fn = pad_right
        elif all_numeric:
            pad_fn = pad_left
        else:
            pad_fn = pad_right
        # second pass does padding and line breaks
        str_cols[col] = [pad_fn(strvalue, colwidth, truncate) for strvalue in strvalues]
    return FormattedColumns(nitems, headers, str_cols, widths)


def format_row(columns: List[str], widths: List[int]) -> str:
    """Format one row of the table. This may go into multiple lines if
    any of the columns were wrapped."""
    row = ""
    had_values = True
    lineno = 0
    while had_values:
        had_values = False
        line = ""
        for (i, c) in enumerate(columns):
            collines = c.split("\n")
            if len(collines) > lineno:
                had_values = True
                line += "| " + collines[lineno] + " "
            else:
                line += "|" + " " * (2 + widths[i])
        if had_values:
            row += line + "|\n"
            lineno += 1
    return row[0:-1]


def row_generator(cols: FormattedColumns, title: Optional[str] = None, nl=False):
    if title is not None:
        for line in title.split("\n"):
            if nl:
                yield line + "\n"
            else:
                yield line
    # header row
    for line in format_row(cols.headers, cols.widths).split("\n"):
        if nl:
            yield line + "\n"
        else:
            yield line
    # divider row
    div = "|" + "|".join(["_" * (width + 2) for width in cols.widths]) + "|"
    if nl:
        yield div + "\n"
    else:
        yield div
    # values
    for i in range(cols.nitems):
        for line in format_row([values[i] for values in cols.columns.values()], cols.widths).split(
            "\n"
        ):
            if nl:
                yield line + "\n"
            else:
                yield line


def print_columns(
    columns: Dict[str, Any],
    precision=-1,
    null_value: str = "None",
    spec: Dict[str, ColSpec] = {},
    paginate: bool = True,
    title: Optional[str] = None,
) -> None:
    """Print the columns as a table.

    A precision of -1 means to adjust based on whether absolute value is less than one.
    """
    cols = format_columns(columns, precision, null_value, spec)
    if paginate:
        click.echo_via_pager(row_generator(cols, title=title, nl=True))
    else:
        for line in row_generator(cols, title=title):
            click.echo(line)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.

import sys
import traceback

import click

from .errors import CalledProcessError, UserAbort, ConfigurationError, InternalError, BatchModeError

from .dws import cli, is_verbose_mode


def main():
    try:
        cli()
        sys.exit(0)
    except CalledProcessError as e:
        click.echo(e.stdout)
        click.echo(e.stderr)
        tb = traceback.format_exc()
        click.echo(tb, err=True)
        click.echo("Error in subprocess call. Command was:\n   %s" % e.cmd, err=True)
        sys.exit(1)
    except UserAbort as e:
        click.echo("Not a positive response, exiting without doing anything", err=True)
        sys.exit(1)
    except ConfigurationError as e:
        if is_verbose_mode():
            tb = traceback.format_exc()
            click.echo(tb, err=True)
        click.echo("Configuration error: " + str(e), err=True)
        sys.exit(1)
    except BatchModeError as e:
        if is_verbose_mode():
            tb = traceback.format_exc()
            click.echo(tb, err=True)
        click.echo("Running in --batch mode, but user input required for %s" % str(e), err=True)
        sys.exit(1)
    except InternalError as e:
        tb = traceback.format_exc()
        click.echo(tb, err=True)
        click.echo("Aborting due to internal error: %s" % str(e), err=True)
        sys.exit(1)
    except Exception as e:
        tb = traceback.format_exc()
        click.echo(tb, err=True)
        click.echo("Aborting due to unexpected exception: %s" % repr(e), err=True)
        sys.exit(1)



"""
Common utils for wrapping objects with the Lineage API.
"""
import datetime
from typing import Optional, Union, cast, Dict
from os.path import exists

from dataworkspaces.workspace import Workspace, ResourceRoles, ResourceRef
from dataworkspaces.utils.lineage_utils import LineageError, infer_step_name
from dataworkspaces.kits.jupyter import get_step_name_for_notebook
from dataworkspaces.lineage import ResultsLineage
from dataworkspaces.resources.api_resource import API_RESOURCE_TYPE, ApiResource
from dataworkspaces.errors import ConfigurationError

import numpy as np # type: ignore

try:
    import pandas # type: ignore
except ImportError:
    pandas = None

try:
    import tensorflow # type: ignore
except ImportError:
    tensorflow = None  # type: ignore


class NotSupportedError(ConfigurationError):
    """Thrown when a wrapper encounters an unsupported configuration.
    """

    pass


def _infer_step_name() -> str:
    """Come up with a step name by looking at whether this is a notebook
    and then the command line arguments.
    """
    try:
        notebook_name = get_step_name_for_notebook()
        if notebook_name is not None:
            return notebook_name
    except:
        pass  # not a notebook
    return infer_step_name()


def _metric_scalar_to_json(v):
    if isinstance(v, int) or isinstance(v, str):
        return v
    elif isinstance(v, np.int64) or isinstance(v, np.int32):
        return int(v)
    elif isinstance(v, np.float64) or isinstance(v, np.float32):
        return float(v)
    elif isinstance(v, datetime.datetime):
        return v.isoformat()
    else:
        return v


def _metric_obj_to_json(v):
    if isinstance(v, dict):
        return {k: _metric_obj_to_json(vi) for (k, vi) in v.items()}
    elif isinstance(v, list) or isinstance(v, tuple):
        return [_metric_obj_to_json(vi) for vi in v]
    else:
        return _metric_scalar_to_json(v)


def _add_to_hash(array_data, hash_state):
    if isinstance(array_data, np.ndarray):
        hash_state.update(array_data.data)
    elif (pandas is not None) and isinstance(array_data, pandas.DataFrame):
        for c in array_data.columns:
            hash_state.update(array_data[c].to_numpy(copy=False).data)
    elif (pandas is not None) and isinstance(array_data, pandas.Series):
        hash_state.update(array_data.to_numpy(copy=False).data)
    elif isinstance(array_data, tuple) or isinstance(array_data, list):
        # Tensorflow frequently puts the parts of a dataset in a tuple.
        # For example: (features, labels)
        for element in array_data:
            _add_to_hash(element, hash_state)
    elif isinstance(array_data, dict):
        # Tensorflow uses a dict (specifically OrderedDict) to store
        # the columns of a CSV.
        for column in array_data.values():
            _add_to_hash(column, hash_state)
    elif (tensorflow is not None) and isinstance(array_data, tensorflow.data.Dataset):  # type: ignore
        # We need to iterate through the dataset, to force an eager evaluation
        for t in array_data:
            _add_to_hash(t, hash_state)
    elif (tensorflow is not None) and isinstance(array_data, tensorflow.Tensor):  # type: ignore
        if hasattr(array_data, "numpy"):
            _add_to_hash(array_data.numpy(), hash_state)
        else:
            raise Exception(
                "Tensor type %s is not in eager mode, cannot convert to numpy, value was: %s"
                % (type(array_data), repr(array_data))
            )
    elif (
        isinstance(array_data, np.uint8)
        or isinstance(array_data, np.int8)
        or isinstance(array_data, np.int32)
        or isinstance(array_data, np.int64)
    ):
        hash_state.update(bytes(int(array_data)))
    else:
        raise Exception(
            "Unable to hash data type %s, data was: %s" % (type(array_data), array_data)
        )


def _find_resource(
    workspace: Workspace, role: str, name_or_ref: Optional[Union[str, ResourceRef]] = None
) -> ResourceRef:
    resource_names = [n for n in workspace.get_resource_names()]
    if isinstance(name_or_ref, str):
        if (
            (not name_or_ref.startswith("./"))
            and (not name_or_ref.startswith("/"))
            and (name_or_ref in resource_names)
        ):
            return ResourceRef(name_or_ref)
        elif exists(name_or_ref):
            return workspace.map_local_path_to_resource(
                name_or_ref, expecting_a_code_resource=False
            )
        else:
            raise LineageError(
                "Could not find a resource for '"
                + name_or_ref
                + "' with role '"
                + role
                + "' in your workspace. Please create a resource"
                + " using the 'dws add' command or correct the name. "
                + "Currently defined resources are: "
                + ", ".join(
                    ["%s (role %s)" % (n, workspace.get_resource_role(n)) for n in resource_names]
                )
                + "."
            )
    elif isinstance(name_or_ref, ResourceRef):
        workspace.validate_resource_name(name_or_ref.name, name_or_ref.subpath)
        return name_or_ref
    else:
        # no resource specified. If we have exactly one for that role,
        # we will use it
        resource_for_role = None
        for rname in workspace.get_resource_names():
            if workspace.get_resource_role(rname) == role:
                if resource_for_role is None:
                    resource_for_role = ResourceRef(rname, subpath=None)
                else:
                    raise LineageError(
                        "There is more than one resource for role "
                        + role
                        + " in your workspace. Please specify the resource you want"
                        + " in model wrapping function or use a wrapped data set"
                    )
        if resource_for_role is not None:
            return resource_for_role
        else:
            raise LineageError(
                "Could not find a "
                + role
                + " resource in your workspace. Please create a resource"
                + " using the dws add command."
            )


class _DwsModelState:
    def __init__(
        self,
        workspace: Workspace,
        input_resource: Optional[Union[str, ResourceRef]] = None,
        results_resource: Optional[Union[str, ResourceRef]] = None,
    ):
        self.workspace = workspace
        self.results_ref = _find_resource(workspace, ResourceRoles.RESULTS, results_resource)
        self.default_input_resource = input_resource
        self.api_resource_cache = {}  # type: Dict[str,ApiResource]
        self.lineage = ResultsLineage(
            _infer_step_name(), datetime.datetime.now(), {}, [], [], self.results_ref, workspace
        )

    def find_input_resources_and_return_if_api(
        self, data, target_data=None
    ) -> Optional[ApiResource]:
        if hasattr(data, "resource"):
            ref = data.resource
        else:
            ref = _find_resource(
                self.workspace, ResourceRoles.SOURCE_DATA_SET, self.default_input_resource
            )
        self.lineage.add_input_ref(ref)
        data_resource_type = self.workspace.get_resource_type(ref.name)
        if target_data is not None and hasattr(target_data, "resource"):
            target_ref = data.resource
            if target_ref != ref:  # only can happen if resource is specified on data
                if (
                    data_resource_type == API_RESOURCE_TYPE
                    or self.workspace.get_resource_type(target_ref.name) == API_RESOURCE_TYPE
                ):
                    raise NotSupportedError(
                        "Currently, we do not support API Resources where the feature and target data are from different resources (%s and %s)."
                        % (ref, target_ref)
                    )
                self.lineage.add_input_ref(target_ref)
        if data_resource_type == API_RESOURCE_TYPE:
            if ref.name not in self.api_resource_cache:
                self.api_resource_cache[ref.name] = cast(
                    ApiResource, self.workspace.get_resource(ref.name)
                )
            return self.api_resource_cache[ref.name]
        else:
            return None

    def write_metrics_and_complete(self, metrics):
        metrics = _metric_obj_to_json(metrics)
        if self.workspace.verbose:
            print("dws>> Metrics: %s" % repr(metrics))
        self.lineage.write_results(metrics)
        self.lineage.complete()

    def reset_lineage(self):
        """If you are rerunning a step, call this to reset the start and execution
        times as well as the in_progress marker in the lineage.
        """
        self.lineage.step.execution_time_seconds = None
        self.lineage.step.start_time = datetime.datetime.now()
        self.lineage.in_progress = True



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
import click
from typing import Optional, cast, Dict, List, Any, Tuple

assert Dict and List and Any  # for pyflakes
from collections import Counter
from os.path import join
import sys

from dataworkspaces.workspace import (
    RESOURCE_ROLE_CHOICES,
    Workspace,
    SnapshotWorkspaceMixin,
    JSONDict,
    ResourceRoles,
    FileResourceMixin,
)
from dataworkspaces.errors import ConfigurationError
from dataworkspaces.utils.print_utils import print_columns, ColSpec
from dataworkspaces.utils.lineage_utils import make_lineage_table


METRIC_NAME_WIDTH = 18
METRIC_VAL_WIDTH = 12

NUM_METRICS = 2


def print_snapshot_history(
    workspace: SnapshotWorkspaceMixin, reverse: bool = True, max_count: Optional[int] = None
):
    history = workspace.list_snapshots(reverse, max_count)
    # find the most common metrics
    mcounter = Counter()  # type: Counter
    for md in history:
        if md.metrics is not None:
            mcounter.update(md.metrics.keys())
    metric_names = [m for (m, cnt) in mcounter.most_common(NUM_METRICS)]
    spec = {
        "Hash": ColSpec(width=8),
        "Tags": ColSpec(width=20),
        "Created": ColSpec(width=19),
        "Message": ColSpec(width=30),
    }
    hashes = []  # type: List[str]
    tags = []  # type: List[str]
    created = []  # type: List[str]
    metrics = {n: [] for n in metric_names}  # type: Dict[str,List[Any]]
    messages = []  # type: List[str]

    returned = 0
    for md in history:
        hashes.append(md.hashval[0:7])
        tags.append(", ".join(md.tags))
        created.append(md.timestamp[0:-7])
        messages.append(md.message)
        for m in metric_names:
            metrics[m].append(md.metrics[m] if md.metrics is not None and m in md.metrics else None)
        returned += 1
    columns = {"Hash": hashes, "Tags": tags, "Created": created}
    for m in metric_names:
        columns[m] = metrics[m]
        spec[m] = ColSpec(width=25, truncate=True)
    columns["Message"] = messages
    print_columns(columns, null_value="", spec=spec, paginate=False, title="History of snapshots")
    if max_count is not None and returned == max_count:
        click.echo("Showing first %d snapshots" % max_count)
    else:
        click.echo("%d snapshots total" % returned)


def print_resource_status(workspace: Workspace):
    names_by_role = {role: [] for role in RESOURCE_ROLE_CHOICES}  # type:Dict[str,List[str]]
    resource_names = []
    roles = []
    types = []
    params = []
    missing_roles = []
    # we are going to order resources by role
    for rname in workspace.get_resource_names():
        role = workspace.get_resource_role(rname)
        names_by_role[role].append(rname)
    for role in RESOURCE_ROLE_CHOICES:
        if len(names_by_role[role]) > 0:
            for rname in names_by_role[role]:
                resource_names.append(rname)
                roles.append(role)
                types.append(workspace.get_resource_type(rname))
                params.append(
                    ",\n".join(
                        [
                            "%s=%s" % (pname, pval)
                            for (pname, pval) in workspace._get_resource_params(rname).items()
                            if pname not in ("resource_type", "name", "role")
                        ]
                    )
                )
        else:
            missing_roles.append(role)
    print_columns(
        {"Resource": resource_names, "Role": roles, "Type": types, "Parameters": params},
        # spec={'Parameters':ColSpec(width=40)},
        null_value="",
        title="Resources for workspace: %s" % workspace.name,
        paginate=False,
    )
    if len(missing_roles) > 0:
        click.echo("No resources for the following roles: %s." % ", ".join(missing_roles))


def report_history_command(workspace: Workspace, limit: Optional[int] = None):
    if not isinstance(workspace, SnapshotWorkspaceMixin):
        raise ConfigurationError(
            "Workspace %s cannot perform snapshots, history not available" % workspace.name
        )
    else:
        print_snapshot_history(
            cast(SnapshotWorkspaceMixin, workspace), reverse=True, max_count=limit
        )


def report_status_command(workspace: Workspace):
    print("Status for workspace: %s" % workspace.name)
    print_resource_status(workspace)


def report_lineage_command(workspace: Workspace, tag_or_hash: Optional[str] = None):
    snapshot_hash = None  # type: Optional[str]
    if tag_or_hash is not None:
        md = workspace.as_snapshot_ws().get_snapshot_by_tag_or_hash(tag_or_hash)
        snapshot_hash = md.hashval
        title = "Lineage for %s" % tag_or_hash
    else:
        title = "Lineage for current state"
    refs = []
    ltypes = []
    details = []
    inputs = []  # type:List[Optional[str]]
    for (ref, ltype, detail, input_list) in make_lineage_table(
        workspace.get_instance(), workspace.as_lineage_ws().get_lineage_store(), snapshot_hash
    ):
        refs.append(ref)
        ltypes.append(ltype)
        details.append(detail)
        if input_list is not None:
            inputs.append("\n".join(input_list))
        else:
            inputs.append(None)
    print_columns(
        {"Resource": refs, "Type": ltypes, "Details": details, "Inputs": inputs},
        title=title,
        paginate=False,
    )


def _find_results_file_if_present(
    workspace: Workspace, subpath: str, resource_name: Optional[str] = None
) -> Optional[Tuple[JSONDict, str]]:
    if resource_name is not None:
        check_resources = [resource_name]
    else:
        check_resources = [
            rn
            for rn in workspace.get_resource_names()
            if workspace.get_resource_role(rn) == ResourceRoles.RESULTS
        ]
    for rn in check_resources:
        resource = workspace.get_resource(rn)
        if not isinstance(resource, FileResourceMixin):
            continue
        if not resource.does_subpath_exist(subpath, must_be_file=True):
            continue
        return (resource.read_results_file(subpath), "%s:/%s" % (rn, subpath))
    return None  # not found


def _get_results(
    workspace: Workspace, tag_or_hash: Optional[str] = None, resource_name: Optional[str] = None
) -> Optional[Tuple[JSONDict, str]]:
    """Get a results file as a parsed json dict. If no resource or snapshot
    is specified, searches all the results resources for a file. If a snapshot
    is specified, we look in the subdirectory where the resuls have been moved.
    If no snapshot is specified, and we don't find a file, we look in the most
    recent snapshot.

    Returns a tuple with the results and the logical path (resource:/subpath) to
    the results. If nothing is found, returns None.
    """
    if tag_or_hash is not None:
        if not isinstance(workspace, SnapshotWorkspaceMixin):
            raise ConfigurationError("Workspace %s does not support snapshots" % workspace.name)
        md = workspace.as_snapshot_ws().get_snapshot_by_tag_or_hash(tag_or_hash)
        subpath = join(md.relative_destination_path, "results.json")
        return _find_results_file_if_present(workspace, subpath, resource_name)
    else:
        result = _find_results_file_if_present(workspace, "results.json", resource_name)
        if result is not None:
            return result
        # not found - ok, try the snapshot
        if not isinstance(workspace, SnapshotWorkspaceMixin):
            return None
        print(
            "Did not find a results.json file in current workspace, checking most recent snapshot...",
            file=sys.stderr,
        )
        rmd = workspace.get_most_recent_snapshot()
        if rmd is not None:
            subpath = join(rmd.relative_destination_path, "results.json")
            return _find_results_file_if_present(workspace, subpath, resource_name)
        else:
            return None


def report_results_command(
    workspace: Workspace, tag_or_hash: Optional[str] = None, resource_name: Optional[str] = None
):
    results = _get_results(workspace, tag_or_hash, resource_name)
    if results is None:
        msg = "Could not find results file"
        if tag_or_hash:
            msg += " for snapshot %s" % tag_or_hash
        if resource_name:
            msg += " for resource %s" % resource_name
        raise ConfigurationError(msg)
    else:
        (data, path) = results
        click.echo("Results file at %s" % path)

        def print_dict_as_table(d, name, serialize_dicts=False):
            keys = []
            values = []
            for (k, v) in d.items():
                if not isinstance(v, dict):
                    keys.append(k)
                    values.append(v)
                elif serialize_dicts:
                    keys.append(k)
                    values.append(", ".join(["%s: %s" % (mk, mv) for (mk, mv) in v.items()]))
            click.echo()
            print_columns({"Key": keys, "Value": values}, paginate=False, title=name)

        print_dict_as_table(data, "General Properties")
        if "parameters" in data:
            print_dict_as_table(data["parameters"], "Parameters")
        if "metrics" in data:
            print_dict_as_table(data["metrics"], "Metrics")
            for (k, v) in data["metrics"].items():
                if isinstance(v, dict):
                    print_dict_as_table(v, "Metrics: " + k, serialize_dicts=True)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.

import click
from typing import Optional, List, cast

from dataworkspaces.errors import ConfigurationError
from dataworkspaces.workspace import (
    Workspace,
    LocalStateResourceMixin,
    SyncedWorkspaceMixin,
    CentralWorkspaceMixin,
    Resource,
)


def build_resource_list(
    workspace: Workspace, only: Optional[List[str]], skip: Optional[List[str]]
) -> List[str]:
    """Build up our resource name list for either push or pull commands.
    """
    if (only is not None) and (skip is not None):
        raise ConfigurationError("Cannot specify both --only and --skip")
    all_resource_names_set = frozenset(workspace.get_resource_names())
    local_state_names_set = frozenset(workspace.get_names_of_resources_with_local_state())
    if only is not None:
        only_set = frozenset(only)
        invalid = only_set.difference(all_resource_names_set)
        if len(invalid) > 0:
            raise ConfigurationError(
                "Invalid resource names were included with --only: %s" % ", ".join(sorted(invalid))
            )
        nonsync_rnames = only_set.difference(local_state_names_set)
        if len(nonsync_rnames) > 0:
            click.echo(
                "Skipping the following resources, which do not have local state: %s"
                % ", ".join(sorted(nonsync_rnames))
            )
        return [rn for rn in only if rn in local_state_names_set]
    elif skip is not None:
        skip_set = frozenset(skip)
        invalid = skip_set.difference(all_resource_names_set)
        if len(invalid) > 0:
            raise ConfigurationError(
                "Invalid resource names were included with --skip: %s" % ", ".join(sorted(invalid))
            )
        nonsync_rnames = all_resource_names_set.difference(skip_set).difference(
            local_state_names_set
        )
        if len(nonsync_rnames) > 0:
            click.echo(
                "Skipping the following resources, which do not have local state: %s"
                % ", ".join(sorted(nonsync_rnames))
            )
        skip_set = skip_set.union(nonsync_rnames)
        return [rn for rn in workspace.get_resource_names() if rn not in skip_set]
    else:
        nonsync_rnames = all_resource_names_set.difference(local_state_names_set)
        if len(nonsync_rnames) > 0:
            click.echo(
                "Skipping the following resources, which do not have local state: %s"
                % ", ".join(sorted(nonsync_rnames))
            )
        return [rn for rn in workspace.get_resource_names() if rn not in nonsync_rnames]


def push_command(
    workspace: Workspace,
    only: Optional[List[str]] = None,
    skip: Optional[List[str]] = None,
    only_workspace: bool = False,
) -> int:
    """Run the push command on the pushable resources and the workspace.
    """
    if only_workspace:
        if isinstance(workspace, CentralWorkspaceMixin):
            raise ConfigurationError(
                "--only-workspace not valid for central workspace %s" % workspace.name
            )
        resource_list = []  # type: List[LocalStateResourceMixin]
    else:
        resource_list = [
            cast(LocalStateResourceMixin, workspace.get_resource(rn))
            for rn in build_resource_list(workspace, only, skip)
        ]

    if isinstance(workspace, CentralWorkspaceMixin):
        if len(resource_list) == 0:
            click.echo("No resources to push.")
            return 0
        else:
            print(
                "Pushing resources: %s" % ", ".join([cast(Resource, r).name for r in resource_list])
            )
            workspace.push_resources(resource_list)
    elif isinstance(workspace, SyncedWorkspaceMixin):
        if len(resource_list) > 0:
            click.echo(
                "Pushing workspace and resources: %s"
                % ", ".join([cast(Resource, r).name for r in resource_list])
            )
        elif not only_workspace:
            click.echo("No resources to push, will still push workspace")
        else:
            click.echo("Pushing workspace.")
        workspace.push(resource_list)

    workspace.save("push command")
    return len(resource_list)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
import os
import tempfile
import hashlib

from typing import Dict, Optional, List, Tuple, Iterable, Callable

assert Dict

from dataworkspaces.utils.subprocess_utils import call_subprocess
from dataworkspaces.utils.git_utils import GIT_EXE_PATH
from dataworkspaces.utils.file_utils import safe_rename

BUF_SIZE = 65536  # read stuff in 64kb chunks


def compute_hash(tmpname: str) -> str:
    sha1 = hashlib.sha1()
    with open(tmpname, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def compute_size(fname: str) -> str:
    if not (os.path.exists(fname)):
        raise Exception("File %s does not exist" % fname)
    statinfo = os.stat(fname)
    return str(statinfo.st_size)


BLOB = "blob"
TREE = "tree"
TYPES = (BLOB, TREE)


class HashEntry:
    def __init__(self, name: str, sha: Optional[str]):
        self.name = name
        self.sha = sha


class HashBlob(HashEntry):
    type = BLOB

    def __init__(self, name: str, sha: str):
        super().__init__(name, sha)


class HashTree(HashEntry):
    type = TREE

    def __init__(self, rootdir: str, name: str, add_to_git: bool = True):
        super().__init__(name, None)
        self.path = rootdir
        self.cache = []  # type: List[Tuple[str,str,str]]
        # JF 2021-07: ache loookups are slow, so store a mapping frm name to index.
        # No longer can delete items from cache, but we dont need that functionality.
        self.by_name = {} # type: Dict[str,int]
        self.add_to_git = add_to_git

    def _index_by_name(self, name):
        return self.by_name.get(name, -1)

    _map_id_to_type = {
        BLOB: HashBlob
        # HashTree is added at the end of the class
    }  # type: Dict[str, Callable]

    def _iter_convert_to_object(self, iterable: Iterable[Tuple[str, str, str]]):
        for sha, mode, name in iterable:
            try:
                yield self._map_id_to_type[mode](sha, mode, name)
            except KeyError:
                raise TypeError("Unknown mode %s found in tree data for path '%s'" % (mode, name))

    def trees(self):
        """return the list of direct subtrees"""
        return [i for i in self if i.type == TREE]

    def blobs(self):
        """return the list of file hashes in this directory"""
        return [i for i in self if i.type == BLOB]

    def add(self, name: str, mode: str, sha: str, force: bool = False):
        """Add elements to a tree. 
           :name: name of the file
           :mode: blob or tree
           :sha:  hash
           :force: if true, overwrite previous entry
        """
        item = (sha, mode, name)
        index = self._index_by_name(name)
        if index == -1:
            # this name is not in the cache
            index = len(self.cache)
            self.by_name[name] = index            
            self.cache.append(item)
            
        else:
            if force:
                self.cache[index] = item
            else:
                prev_item = self.cache[index]
                if prev_item[0] != sha or prev_item[1] != mode:
                    raise ValueError("Item %r existed with different properties" % name)


    def sort(self):
        """sort the cache in alphabetical order"""
        self.cache.sort(key=(lambda t: t[2]))

    def write(self):
        self.sort()
        # write to a temp file
        fd, tmpname = tempfile.mkstemp()
        with os.fdopen(fd, "w") as f:
            for (mode, sha, name) in self.cache:
                f.write("{}\t{}\t{}\n".format(mode, sha, name))
        # hash the file
        self.hash = compute_hash(tmpname)
        # that is the name of the file
        objfile = os.path.join(self.path, self.hash)
        os.chmod(tmpname, int("755", 8))
        safe_rename(tmpname, objfile)
        if self.add_to_git:
            call_subprocess([GIT_EXE_PATH, "add", self.hash], cwd=self.path, verbose=False)
        return self.hash

    # List protocol
    def __getslice__(self, i, j):
        return list(self._iter_convert_to_object(self.cache[i:j]))

    def __iter__(self):
        return self._iter_convert_to_object(self.cache)

    def __len__(self):
        return len(self.cache)

    def __getitem__(self, item):
        if isinstance(item, int):
            info = self.cache[item]
            return self._map_id_to_type[info[1]](info[0], info[1], info[2])
        raise TypeError("Invalid index type: %r" % item)

    def __contains__(self, item):
        if isinstance(item, HashEntry):
            for info in self.cache:
                if item.sha == info[0]:
                    return True
        return False

    def __reversed__(self):
        return reversed(self._iter_convert_to_object(self.cache))


HashTree._map_id_to_type[TREE] = HashTree


def generate_hashes(
    path_where_hashes_are_stored: str,
    local_dir: str,
    ignore: List[str] = [],
    hash_fun: Callable[[str], str] = compute_hash,
    add_to_git: bool = True,
    verbose: bool = False,
) -> str:
    """traverse a directory tree rooted at :local_dir: and construct the tree hashes
       in the directory :path_where_hashes_are_stored:
       skip directories in :ignore:"""
    hashtbl = {}  # type: Dict[str, str]
    for root, dirs, files in os.walk(local_dir, topdown=False):
        if os.path.basename(root) in ignore:
            if verbose:
                print("skipping %s" % root)
            continue
        if verbose:
            print("generate_hashes: walk at %s" % root)
            print("  files: %s" % ", ".join(files))
            print("  dirs: %s" % ", ".join(dirs))
        t = HashTree(path_where_hashes_are_stored, root, add_to_git=add_to_git)
        for f in files:
            # print(f)
            sha = hash_fun(os.path.join(root, f))
            t.add(f, BLOB, sha)
        for dir in dirs:
            # print(dir)
            if dir in ignore:
                if verbose:
                    print("skipping dir %s under %s" % (dir, root))
                continue
            dirsha = hashtbl[os.path.join(root, dir)]
            t.add(dir, TREE, dirsha)
        h = t.write()
        hashtbl[root] = h
    return hashtbl[local_dir].strip()


def _get_next_element(
    dl: List[str], startindex: int, ignore: List[str], verbose: bool = False
) -> Tuple[Optional[str], int]:
    index = startindex
    for d in dl[startindex:]:
        index = index + 1
        if d in ignore:
            if verbose:
                print("_get_next_element: skipping %s" % d)
            continue
        return d, index
    return None, -1


def check_hashes(
    roothash: str,
    basedir_where_hashes_are_stored: str,
    local_dir: str,
    ignore: List[str] = [],
    hash_fun: Callable[[str], str] = compute_hash,
    verbose: bool = False,
) -> bool:
    """Traverse a directory tree rooted at :local_dir: and check that the files
       match the hashes kept in :basedir_where_hashes_are_stored: and that no new
       files have been added.
       Ignore directories in :ignore:"""
    hashfile = os.path.abspath(os.path.join(basedir_where_hashes_are_stored, roothash))
    if verbose:
        print("Checking hashes. Root hash ", roothash, " root hashfile ", hashfile)

    hashtbl = {os.path.abspath(local_dir): hashfile}
    for root, dirs, files in os.walk(local_dir, topdown=True):
        if verbose:
            print("check_hashes: walk at root=%s" % root)
            print("  files: %s" % ", ".join(files))
            print("  dirs: %s" % ", ".join(dirs))
        if os.path.basename(root) in ignore:
            if verbose:
                print("ignoring %s" % root)
            continue
        try:
            try:
                hashfile = hashtbl[os.path.abspath(root)]
            except KeyError as k:
                print("Key Error:", k)
                print("Hashtbl is\n", hashtbl)
                assert 0, "key error in hash table at %s" % root
            try:
                fd = open(hashfile, "r")
            except:
                print("File %s not found or not readable" % hashfile)
                return False

            dirs.sort()
            files.sort()

            f_index = 0
            d_index = 0
            for line in fd.readlines():
                line = line.rstrip("\n")
                h, kind, name = line.split("\t")
                # print("Line from hash file ", h, kind, name)
                if kind == BLOB:
                    f, f_index = _get_next_element(files, f_index, ignore, verbose=verbose)
                    if verbose:
                        print("f = ", f, " name =", name)
                    if f != name or (f is None):
                        print("File mismatch:", f, " and (hash says)", name)
                        return False
                    sha = hash_fun(os.path.join(root, f))
                    if sha != h:
                        print("Hash mismatch for file: ", f, ":", sha, " and (hash says)", h)
                        return False
                elif kind == TREE:
                    d, d_index = _get_next_element(dirs, d_index, ignore, verbose=verbose)
                    if verbose:
                        print("d = ", d, " name = ", name)
                    if (d != name) or (d is None):
                        print("Dir mismatch: ", d, " and (hash says)", name)
                        return False
                    hashtbl[os.path.abspath(os.path.join(root, d))] = os.path.join(
                        basedir_where_hashes_are_stored, h
                    )
                else:
                    assert kind == TREE or kind == BLOB
            # Make sure that there are no extra files at end of list
            f, f_index = _get_next_element(files, f_index, ignore, verbose=verbose)
            if f is not None:
                print(
                    "Hash mismatch for file %s in directory %s: extra file in directory, not in previous hash"
                    % (f, root)
                )
                return False
            d, d_index = _get_next_element(dirs, d_index, ignore, verbose=verbose)
            if d is not None:
                print(
                    "Hash mismatch for subdirectory %s in directory %s: extra subdirectory, not in previous hash"
                    % (d, root)
                )
                return False
        finally:
            fd.close()
    return True


def generate_sha_signature(
    rsrcdir: str, localpath: str, ignore: List[str] = [], verbose: bool = False
) -> str:
    return generate_hashes(
        rsrcdir, localpath, ignore=ignore, hash_fun=compute_hash, verbose=verbose
    )


def check_sha_signature(
    hashval: str, rsrdir: str, localpath: str, ignore: List[str] = [], verbose: bool = False
) -> bool:
    return check_hashes(
        hashval, rsrdir, localpath, ignore=ignore, hash_fun=compute_hash, verbose=verbose
    )


def generate_size_signature(
    rsrcdir: str, localpath: str, ignore: List[str] = [], verbose: bool = False
) -> str:
    return generate_hashes(
        rsrcdir, localpath, ignore=ignore, hash_fun=compute_size, verbose=verbose
    )


def check_size_signature(
    hashval: str, rsrdir: str, localpath: str, ignore: List[str] = [], verbose: bool = False
) -> bool:
    return check_hashes(
        hashval, rsrdir, localpath, ignore=ignore, hash_fun=compute_size, verbose=verbose
    )


# Tests have been moved to tests/test_hashtree.py



import os
from os.path import curdir, abspath, expanduser
from typing import Optional
import click

from dataworkspaces.workspace import Workspace
import dataworkspaces.backends.git as git_backend
from dataworkspaces.utils.git_utils import get_git_config_param, get_remote_origin_url
from dataworkspaces.utils.subprocess_utils import call_subprocess
from dataworkspaces.errors import ConfigurationError

R2D_IMPORT_ERROR="Unable to import repo2docker package. You may need to reinstall Data Workspaces with the 'docker' option:"+\
                 " pip install dataworkspaces[docker]"

def deploy_build_command(
    workspace: Workspace,
    image_name: Optional[str],
    force_rebuild: bool,
    git_user_email: Optional[str],
    git_user_name: Optional[str],
) -> None:
    try:
        from repo2docker.__main__ import make_r2d  # type: ignore
    except ImportError as e:
        raise ConfigurationError(R2D_IMPORT_ERROR) from e
    target_repo_dir = "/home/%s/%s" % (os.environ["USER"], workspace.name)
    if image_name is None:
        image_name = workspace.name
    argv = ["--target-repo-dir", target_repo_dir, "--image-name", image_name, "--no-run"]
    if isinstance(workspace, git_backend.Workspace):
        workspace_dir = workspace.get_workspace_local_path_if_any()
        assert workspace_dir is not None
        user_email = (
            git_user_email
            if git_user_email
            else get_git_config_param(workspace_dir, "user.email", verbose=workspace.verbose)
        )
        user_name = (
            git_user_name
            if git_user_name
            else get_git_config_param(workspace_dir, "user.name", verbose=workspace.verbose)
        )
        argv.append(
            '--appendix=RUN git config --global user.email %s; git config --global user.name "%s"'
            % (user_email, user_name)
        )
        argv.append("dws+" + get_remote_origin_url(workspace_dir, verbose=workspace.verbose))
    else:
        # need to figure out how the clone url works for a non-git workspace
        assert 0, "build not yet implemented for non-git workspaces"

    if force_rebuild:
        click.echo("Forcing remove of image %s." % image_name)
        call_subprocess(
            ["docker", "image", "rm", "-f", "--no-prune", image_name],
            cwd=curdir,
            verbose=workspace.verbose,
        )
    if workspace.verbose:
        click.echo("Command args for repo2docker: %s" % repr(argv))
    r2d = make_r2d(argv=argv)
    r2d.initialize()
    r2d.start()
    click.echo("Build of image %s was successful." % image_name)


def deploy_run_command(
    workspace: Workspace, image_name: Optional[str], no_mount_ssh_keys: bool
) -> None:
    try:
        from repo2docker.__main__ import make_r2d  # type: ignore
    except ImportError as e:
        raise ConfigurationError(R2D_IMPORT_ERROR) from e
    target_repo_dir = "/home/%s/%s" % (os.environ["USER"], workspace.name)
    if image_name is None:
        image_name = workspace.name
    argv = [
        "--target-repo-dir",
        target_repo_dir,
        "--image-name",
        image_name,
    ]
    if not no_mount_ssh_keys:
        dot_ssh = abspath(expanduser("~/.ssh"))
        argv.append("-v")
        argv.append("%s:/home/%s/.ssh" % (dot_ssh, os.environ["USER"]))
    if isinstance(workspace, git_backend.Workspace):
        workspace_dir = workspace.get_workspace_local_path_if_any()
        assert workspace_dir is not None
        argv.append("dws+" + get_remote_origin_url(workspace_dir, verbose=workspace.verbose))
    else:
        # need to figure out how the clone url works for a non-git workspace
        assert 0, "run build not yet implemented for non-git workspaces"
    if workspace.verbose:
        click.echo("Command args for repo2docker: %s" % repr(argv))
    r2d = make_r2d(argv=argv)
    r2d.initialize()
    r2d.run_image()
    click.echo("Run of image %s was successful." % image_name)



# Copyright 2018,2019 by MPI-SWS and Data-ken Research. Licensed under Apache 2.0. See LICENSE.txt.
"""
Error class defintions
"""

from subprocess import CalledProcessError


class ConfigurationError(Exception):
    """Thrown when something is wrong with the system environment.
    """

    pass


class PathNotAResourceError(ConfigurationError):
    """Raised when a path on the local filesystem does not correspond to a
    resource and it is expected to.
    """

    pass


class LineageError(ConfigurationError):
    """An error related to data lineage (e.g. action would make it impossible
    to track lineage)
    """

    pass


class NotSupportedError(ConfigurationError):
    """Raised when a requested capability is not supported by
    the specified resource or workspace.
    """

    pass


class UserAbort(Exception):
    """Thrown when the user requests not to perform the action.
    """

    pass


class InternalError(Exception):
    """Thrown when something unexpected happens.
    """

    pass


class BatchModeError(Exception):
    """Thrown when running in batch mode but user input is required"""

    pass


class ApiParamError(Exception):
    """Incorrect value for an API parameter"""

    pass

class PathError(ApiParamError):
    """A path provided to an API by the user is not valid"""
    pass




import boto3 # type: ignore
from multiprocessing import Process, JoinableQueue, Queue, cpu_count
import time
import json
import argparse
import sys
import gzip
from os.path import join

from dataworkspaces.utils.hash_utils import hash_bytes

CPU_COUNT=cpu_count()

SNAPSHOTS_SUBDIR=".snapshots"
SNAPSHOTS_PREFIX=".snapshots/" # when getting the list of prefixes from s3, there is a trailing slash

class VersionWorker(Process):
    def __init__(self, work_q, result_q, bucket, max_keys, max_depth):
        super().__init__()
        self.work_q = work_q
        self.result_q = result_q
        self.bucket = bucket
        self.max_keys = max_keys
        self.max_depth = max_depth
        self.versions = {}
        self.client = None

    def get_at_prefix(self, prefix, depth):
        next_key_marker = None
        next_version_id_marker = None
        next_depth = depth + 1
        
        while True:
            kwargs = {'Bucket':self.bucket,
                      'MaxKeys':self.max_keys,
                      'Delimiter':'/',
                      'Prefix':prefix}
            if next_key_marker is not None:
                kwargs['KeyMarker'] = next_key_marker
            if next_version_id_marker is not None:
                kwargs['VersionIdMarker'] = next_version_id_marker
            assert self.client is not None
            resp = self.client.list_object_versions(**kwargs)

            if 'CommonPrefixes'  in resp:
                if next_depth>self.max_depth:
                    for subprefix_dict in resp['CommonPrefixes']:
                        subprefix = subprefix_dict['Prefix']
                        if subprefix==SNAPSHOTS_PREFIX:
                            continue
                        self.get_at_prefix(subprefix, next_depth)
                else:
                    for subprefix_dict in resp['CommonPrefixes']:
                        subprefix = subprefix_dict['Prefix']
                        if subprefix==SNAPSHOTS_PREFIX:
                            continue                        
                        self.work_q.put((subprefix, next_depth))

            # get the files
            if 'Versions' in resp:
                entries = resp['Versions']
                for entry in entries:
                    key = entry['Key']
                    if entry['IsLatest'] and key!=prefix:
                        # self.versions[key] = (entry['LastModified'].isoformat(),
                        #                       entry['VersionId'])
                        self.versions[key] = entry['VersionId']
            truncated = resp['IsTruncated']
            if truncated:
                next_key_marker = resp['NextKeyMarker']
                #new_version_id_marker = resp['NextVersionIdMarker']
            else:
                break

    def run(self):
        self.client = boto3.client('s3')
        while True:
            prefix, depth = self.work_q.get()
            if prefix==None:
                self.result_q.put(self.versions)
                break # stop the thread
            self.get_at_prefix(prefix, depth)
            self.work_q.task_done()

def snapshot_multiprocess(bucket, snapshot_dir, max_keys=1000, num_workers=CPU_COUNT, max_depth=2, dry_run=False):
    """Compute the snapshot and store as a hash in the specified directory.
    The filename will be HASH.json.gz. The hashing occurs before compresssing.
    Returns the hashcode and the versions directory."""
    start = time.time()
    work_q = JoinableQueue() # type: JoinableQueue
    work_q.put(('', 0))
    result_q = Queue() # type: Queue
    workers = [VersionWorker(work_q, result_q, bucket, max_keys, max_depth) for i in range(num_workers)]
    for worker in workers:
        worker.start()
    work_q.join()
    for worker in workers:
        work_q.put((None,None),)

    # combine and write the result
    pre_write= time.time()
    versions = {}
    for i in range(num_workers):
        versions.update(result_q.get())
    #sorted_versions = dict(sorted(versions.items()))
    #with open(snapshot_file, 'w') as f:
    #    json.dump(sorted_versions, f, indent=2)
    data = json.dumps(versions, sort_keys=True, indent=2).encode('utf-8')
    hashcode = hash_bytes(data)
    local_file = join(snapshot_dir, hashcode+'.json.gz')
    with open(local_file, 'wb') as f:
        f.write(gzip.compress(data))
    if not dry_run:
        import s3fs # type: ignore
        fs = s3fs.S3FileSystem()
        snapshot_path = join(join(bucket, SNAPSHOTS_SUBDIR), f'{hashcode}.json.gz')
        fs.put(local_file, snapshot_path)
        print(f"Uploaded snapshot to s3://{snapshot_path}")
    end = time.time()
    print(f"Completed snapshot of {len(versions)} objects in {round(end-start, 1)} seconds")
    print(f"Time to write (included in total) was {round(end-pre_write, 2)} seconds")
    print(f"hashcode={hashcode}")
    return (hashcode, versions)

# just for testing
def main(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser()
    parser.add_argument('--workers', default=CPU_COUNT, type=int,
                        help=f"Number of workers, defaults to number of CPUs ({CPU_COUNT})")
    parser.add_argument('--max-keys', default=1000, type=int,
                        help="Maximum keys per request, defaults to 1000")
    parser.add_argument('--max-depth', default=2, type=int,
                        help="Maximum depth before switching to same thread, depth-first.")
    parser.add_argument('--dry-run', default=False, action='store_true',
                        help="If specified, don't upload the snapshot file to the bucket")
    parser.add_argument('bucket', metavar="BUCKET", help="Name of bucket")
    parser.add_argument('snapshot_dir', metavar="SNAPSHOT_DIR",
                        help="Name of directory to store snapshot file")
    args = parser.parse_args(argv)
    snapshot_multiprocess(args.bucket, args.snapshot_dir, num_workers=args.workers,
                          max_keys=args.max_keys, max_depth=args.max_depth, dry_run=args.dry_run)
    return 0

if __name__=='__main__':
    sys.exit(main())
